def write(filename, payload, verbosity=1, **parameters):
    """Write an object to a filename."""
    from ase import io
    from ase import atoms as aseatoms

    # import mcol
    # import mout
    import mrich
    import json
    from .system import System
    from .group import AtomGroup
    import plotly.graph_objects as go
    import pickle

    filename = str(filename)

    if verbosity > 0:
        mrich.writing(filename)

    # Different behaviour depending on format:
    try:
        # pickle:
        if filename.endswith(".pickle"):
            with open(filename, "wb") as f:
                pickle.dump(payload, f)

        # JSON:
        elif filename.endswith(".json"):
            with open(filename, "wt") as f:
                json.dump(payload, f, indent=2)

        # CJSON:
        elif filename.endswith(".cjson"):
            writeCJSON(filename, payload)
        # PBD and amp.System:
        elif (filename.endswith(".pdb") or filename.endswith(".pdb2")) and isinstance(
            payload, System
        ):
            writePDB(filename, payload, verbosity=verbosity - 1)
        elif (filename.endswith(".pdb") or filename.endswith(".pdb2")) and isinstance(
            payload, AtomGroup
        ):
            writePDB(filename, payload, verbosity=verbosity - 1)
        # GRO and amp.System:
        elif filename.endswith(".gro") and isinstance(payload, System):
            writeGRO(filename, payload, verbosity=verbosity - 1)
        # PDB and list of amp.System's:
        elif filename.endswith(".pdb") and isinstance(payload, list):

            ### List of Systems
            if all([isinstance(i, System) for i in payload]):

                for index, frame in enumerate(payload):
                    if index == 0:
                        writePDB(filename, frame, verbosity=verbosity - 1)
                    else:
                        writePDB(
                            filename,
                            frame,
                            verbosity=verbosity - 1,
                            append=True,
                            model=index + 1,
                        )

            elif all([isinstance(i, aseatoms.Atoms) for i in payload]):
                io.write(filename, payload, **parameters)
            else:
                mrich.error("Unsupported")
                raise NotImplementedError

        elif isinstance(payload, System):
            mrich.error(
                f"Filetype {mcol.file}{filename.split('.')[-1]}{mcol.error} not supported for mp.System object",
            )
            return None

        elif isinstance(payload, go.Figure):
            if filename.endswith(".html"):
                payload.write_html(filename, **parameters)
            else:
                if filename.endswith(".pdf"):
                    import plotly.io as pio

                    if pio.kaleido.scope is not None:
                        pio.kaleido.scope.mathjax = None
                payload.write_image(filename, **parameters)

        # Others:
        else:
            io.write(filename, payload, **parameters)
    except TypeError as e:
        mrich.error("Fail.")
        mrich.error(e)
        mrich.error(type(payload), "could not be written to ", filename)
        return None


def read(
    filename, index=None, verbosity=1, tagging=True, tagByResidue=False, **parameters
):
    """Wrapper for ase.io.read"""
    from ase import io
    import mcol
    import mout

    if verbosity > 0:
        mout.out(
            "reading " + mcol.file + filename + mcol.clear + " ... ", end=""
        )  # user output

    try:
        atoms = io.read(filename, index, **parameters)
    except FileNotFoundError:
        mout.out("Fail.")
        mout.errorOut(
            "Could not read missing file " + mcol.file + filename,
            code="amp.io.read[1]",
            end="",
        )
        return None

    if tagging and filename.endswith(".pdb"):
        getAndSetTags(filename, atoms, byResidue=tagByResidue)

    if verbosity > 0:
        mout.out("Done.")  # user output

    return atoms


def getAndSetTags(pdb, atoms, byResidue=False):
    from ase import atoms as aseatoms
    import mout

    # Parse the first model in the PDB to get the tags
    taglist = []
    with open(pdb, "r") as input_pdb:
        searching = True
        for line in input_pdb:
            if searching:
                if line.startswith("MODEL"):
                    searching = False
                elif line.startswith("ATOM"):
                    searching = False
                    taglist.append(tagFromLine(line, byResidue=byResidue))
            else:
                if line.startswith("ENDMDL"):
                    break
                if line.startswith("END"):
                    break
                if line.startswith("TER"):
                    continue
                else:
                    # get the tags:
                    taglist.append(tagFromLine(line, byResidue=byResidue))

    # set the tags
    if isinstance(atoms, aseatoms.Atoms):
        for index, tag in enumerate(taglist):
            atoms[index].tag = tag
    else:
        for atms in atoms:
            for index, tag in enumerate(taglist):
                atms[index].tag = tag


def tagFromLine(line, byResidue):
    try:
        if byResidue:
            return int(line[24:27])
        else:
            return int("".join(filter(lambda i: i.isdigit(), line.strip().split()[2])))
    except:
        return 0


def parse(file, verbosity=1, **kwargs):
    """Parse a molecular structure file.

    Supported filetypes
    -------------------

    * PDB: *.pdb
    * Gromacs: *.gro
    * XYZ: *.xyz
    * mol: *.mol
    * Gromacs Index: *.ndx
    * JSON: *.json
    """
    from .ndx import parseNDX

    file = str(file)
    extension = file.split(".")[-1]
    if extension == "pdb":
        return parsePDB(file, verbosity=verbosity, **kwargs)
    elif extension == "gro":
        return parseGRO(file, verbosity=verbosity, **kwargs)
    elif extension == "xyz":
        return parseXYZ(file, verbosity=verbosity, **kwargs)
    elif extension == "mol":
        return parseMol(file, verbosity=verbosity, **kwargs)
    elif extension == "ndx":
        return parseNDX(file, verbosity=verbosity, **kwargs)
    elif extension == "cif":
        return parseCIF(file, verbosity=verbosity, **kwargs)
    elif extension == "json":
        from json import load

        return load(open(file, "rt"))
    else:
        import mout

        mout.errorOut("Unsupported file type for MolParse parsing, using ASE.io.read")
        return read(file, verbosity=verbosity)


def parseMol(mol, verbosity=1):
    """Parse an RDKit Molecule"""
    mol = str(mol)
    assert mol.endswith(".mol"), f"filename doesn't end with '.mol': {mol}"

    if verbosity:
        import mout
        import mcol

        mout.out(
            "parsing " + mcol.file + mol + mcol.clear + " ... ", end=""
        )  # user output

    from rdkit.Chem import MolFromMolFile

    mol = MolFromMolFile(mol)

    if verbosity:
        mout.out("Done.")  # user output

    return mol


def parsePDB(
    pdb,
    systemName=None,
    index=1,
    fix_indices=True,
    fix_atomnames=True,
    autoname_chains=False,
    prune_alternative_sites=False,
    verbosity=1,
    debug=False,
    dry=False,
    resfilter=None,
    alternative_site_warnings=False,
    split_chains_by_type=True,
    remove_empty_chains=True,
    num_appended_hydrogen_chains=False,
    detect_misordered_chains=True,
    reordering_warnings=False,
    reordering_summary_warnings=True,
    skipping_residue_warning=True,
    keep_headers=False,
):
    """Parse a PDB"""
    if resfilter is None:
        resfilter = []

    try:

        from .system import System
        from .chain import Chain
        from .residue import Residue, RES_TYPES
        import mout
        import mcol

        pdb = str(pdb)
        assert pdb.endswith(".pdb"), f"filename doesn't end with '.pdb': {pdb}"

        try:
            index = int(index)
        except:
            if index == ":":
                if verbosity > 0:
                    mout.warningOut("Parsing all models in " + mcol.file + pdb)
            else:
                mout.errorOut("Unsupported index: '" + str(index) + "'", fatal=True)

        if index == ":":

            all_systems = []

            import subprocess

            last_model_line = subprocess.check_output(
                "grep MODEL " + pdb + " | tail -n1 ", shell=True
            )

            for i in range(1, int(last_model_line.split()[-1]) + 1):

                if verbosity > 0:
                    mout.out(
                        "\rparsing "
                        + mcol.file
                        + pdb
                        + mcol.clear
                        + " (model "
                        + str(i)
                        + ") ... ",
                        end="",
                    )  # user output

                all_systems.append(
                    parsePDB(
                        pdb,
                        systemName=systemName,
                        index=i,
                        fix_indices=fix_indices,
                        fix_atomnames=fix_atomnames,
                        verbosity=verbosity - 1,
                        debug=debug,
                        dry=dry,
                    )
                )

            if verbosity > 0:
                mout.out(
                    "\rparsing "
                    + mcol.file
                    + pdb
                    + mcol.clear
                    + " (models 1-"
                    + str(i)
                    + ") ... Done."
                )  # user output

            return all_systems

        else:

            if verbosity > 0:
                mout.out(
                    "parsing " + mcol.file + pdb + mcol.clear + " ... ", end=""
                )  # user output

            if debug:
                mout.warningOut("Debug mode active!")

            import os

            residue = None
            chain = None
            last_residue_name = None
            last_residue_number = None
            last_chain_name = None
            last_residue_type = None
            res_counter = 0
            chain_counter = 0
            atom_counter = 1
            was_terminal = False

            if systemName is None:
                systemName = os.path.splitext(pdb)[0]
            system = System(name=systemName)

            try:
                with open(pdb, "r") as input_pdb:
                    header_data = []
                    searching = True
                    for line in input_pdb:

                        if keep_headers:
                            if line.startswith("TITLE"):
                                header_data.append(line)
                            elif line.startswith("HEADER"):
                                header_data.append(line)
                            elif line.startswith("COMPND"):
                                header_data.append(line)
                            elif line.startswith("REMARK"):
                                header_data.append(line)
                            elif line.startswith("CRYST"):
                                header_data.append(line)
                            elif line.startswith("SCALE"):
                                header_data.append(line)
                            elif line.startswith("HET   "):
                                header_data.append(line)
                            elif line.startswith("FORMUL"):
                                header_data.append(line)
                            elif line.startswith("HELIX"):
                                header_data.append(line)
                            elif line.startswith("HELIX"):
                                header_data.append(line)
                            elif line.startswith("DBREF"):
                                header_data.append(line)
                            elif line.startswith("SEQADV"):
                                header_data.append(line)
                            elif line.startswith("SEQRES"):
                                header_data.append(line)
                            elif line.startswith("SHEET"):
                                header_data.append(line)
                            elif line.startswith("SITE"):
                                header_data.append(line)
                            elif line.startswith("ORIGX"):
                                header_data.append(line)
                            elif line.startswith("SCALE"):
                                header_data.append(line)
                            elif line.startswith("HETNAM"):
                                header_data.append(line)
                            elif line.startswith("SSBOND"):
                                parsePDBSSBOND(system, line)

                        else:

                            if line.startswith("TITLE"):
                                system.title = line.removeprefix("TITLE ").strip()
                                # mout.debug(f'{system.title=}')
                                continue
                            elif line.startswith("HEADER"):
                                system.header = line.removeprefix("HEADER").strip()
                                # mout.debug(f'{system.header=}')
                                continue

                        if searching:
                            if line.startswith("MODEL"):

                                if index == 1:
                                    searching = False
                                elif " " + str(index) in line:
                                    searching = False

                            elif (
                                index == 1 and line[0:6] in ["ATOM  ", "HETATM"]
                            ) or not searching:
                                searching = False

                                if debug:
                                    mout.debug(line)

                                #### PARSELINE
                                try:
                                    atom = parsePDBAtomLine(
                                        line,
                                        res_counter,
                                        atom_counter,
                                        chain_counter,
                                        debug=debug,
                                        alternative_site_warnings=alternative_site_warnings,
                                    )
                                except Exception as e:
                                    mout.error(f"{pdb=} {index=}")
                                    raise Exception(e)

                                chain = Chain(atom.chain)
                                residue = new_residue(
                                    atom.residue,
                                    res_counter,
                                    atom.res_number,
                                    atom.chain,
                                )
                                residue.addAtom(atom, copy=False)
                                last_residue_name = atom.residue
                                last_residue_number = atom.res_number
                                last_chain_name = atom.chain
                        else:
                            if line.startswith("ENDMDL") and not searching:
                                break
                            elif line.startswith("END") and not searching:
                                break
                            elif (
                                line.startswith(
                                    "# All scores below are weighted scores, not raw scores."
                                )
                                and not searching
                            ):
                                break
                            elif line.startswith("TER"):
                                if verbosity > 1:
                                    mout.warningOut(
                                        "Terminal added to "
                                        + mcol.arg
                                        + chain.name
                                        + ":"
                                        + residue.name
                                    )
                                residue.atoms[-1].terminal = True
                                was_terminal = True
                                continue
                            elif line.startswith("CONECT"):
                                continue
                            elif line.startswith("OMPND"):
                                continue
                            elif line.startswith("COMPND"):
                                continue
                            elif line.startswith("CRYST"):
                                continue
                            elif line.startswith("AUTHO"):
                                continue
                            elif line.startswith("MASTER"):
                                continue
                            elif line.startswith("ANISOU"):
                                continue
                            elif dry and any(res in line for res in RES_TYPES["SOL"]):
                                continue
                            else:

                                if debug:
                                    mout.debug(line)

                                ### PARSELINE
                                try:
                                    atom = parsePDBAtomLine(
                                        line,
                                        res_counter,
                                        atom_counter,
                                        chain_counter,
                                        debug=debug,
                                        alternative_site_warnings=alternative_site_warnings,
                                    )
                                except Exception as e:
                                    mout.error(f"{pdb=} {index=}")
                                    raise Exception(e)

                                make_new_res = False
                                if residue is None:
                                    make_new_res = True
                                if last_residue_name != atom.residue:
                                    make_new_res = True
                                if last_residue_number != atom.res_number:
                                    make_new_res = True
                                if last_chain_name != atom.chain:
                                    make_new_res = True

                                make_new_chain = False
                                if residue is None:
                                    make_new_chain = True
                                if last_chain_name != atom.chain:
                                    make_new_chain = True
                                if was_terminal:
                                    make_new_chain = True

                                if make_new_res:
                                    if residue is not None:
                                        if not resfilter or residue.name in resfilter:
                                            chain.add_residue(residue)
                                            res_counter = res_counter + 1
                                        elif skipping_residue_warning:
                                            mout.warning(
                                                f"Skipping residue {residue} {residue.number}"
                                            )
                                    residue = new_residue(
                                        atom.residue,
                                        res_counter,
                                        atom.res_number,
                                        atom.chain,
                                    )
                                    if (
                                        split_chains_by_type
                                        and last_residue_type != residue.type
                                    ):
                                        make_new_chain = True
                                    if make_new_chain:
                                        if chain is not None:
                                            system.add_chain(chain)
                                            chain_counter += 1
                                        chain = Chain(atom.chain)

                                residue.addAtom(atom, copy=False)

                                atom_counter += 1

                                last_residue_type = residue.type
                                last_residue_number = atom.res_number
                                last_residue_name = atom.residue
                                last_chain_name = atom.chain
                                was_terminal = False

            except FileNotFoundError:
                mout.errorOut(
                    "File " + mcol.file + pdb + mcol.error + " not found", fatal=True
                )

            if keep_headers:
                system._header_data = header_data

            if not resfilter or residue.name in resfilter:
                chain.add_residue(residue)
            elif skipping_residue_warning:
                mout.warning(f"Skipping residue {residue} {residue.number}")

            system.add_chain(chain)

            if remove_empty_chains:
                system.chains = [c for c in system.chains if c.residues]

            if prune_alternative_sites:
                system.prune_alternative_sites()

            if fix_atomnames:
                system.fix_atomnames(verbosity=1)

            if autoname_chains:
                system.autoname_chains()

            if num_appended_hydrogen_chains:

                if reordering_summary_warnings:
                    mout.warning(
                        f"Will try to re-order hydrogens in last {num_appended_hydrogen_chains} chains"
                    )

                assert num_appended_hydrogen_chains < len(system.chains), mout.error(
                    "num_appended_hydrogen_chains >= len(system.chains)"
                )

                hydrogen_chains = []
                re_add_chains = []
                for i in range(num_appended_hydrogen_chains):

                    chain = system.chains.pop(-1)

                    if any([a.symbol != "H" for a in chain.atoms]):
                        re_add_chains.append(chain)
                        continue

                    if reordering_summary_warnings:
                        mout.warning(f"Re-ordering hydrogens in {chain=}")
                    hydrogen_chains.append(chain)

                for chain in re_add_chains:
                    system.add_chain(chain)

                for h_chain in hydrogen_chains:
                    for atom in h_chain.atoms:

                        # chain = system.get_chain(atom.chain)
                        chains = [c for c in system.chains if c.name == atom.chain]
                        for chain in chains:
                            residue = chain.residues[f"n{atom.res_number}"]
                            if residue is not None:
                                break

                        if reordering_warnings:
                            mout.warning(
                                f"Moving {atom} {atom.number} (i) to {residue} {residue.number} (res_indices {atom.res_index}-->{residue.index})"
                            )

                        if residue is not None:
                            residue.add_atom(atom)
                        else:
                            mout.warning(
                                f"Didn't find a home for {atom} {atom.number} ({pdb=})"
                            )
                            mout.warning(f"Skipping what appears to be a normal chain")
                            system.add_chain(h_chain)
                            break

            if detect_misordered_chains:
                for chain in system.chains:
                    for i, res in enumerate(chain.residues):

                        if res.number in [r.number for r in chain.residues[:i]]:
                            mout.warning(
                                f"{res} {res.number} has already appeared in chain {chain} {chain.index}"
                            )

            if fix_indices:
                system.fix_indices()

            if verbosity > 0:
                mout.out("Done.")  # user output

            return system

    except KeyboardInterrupt:
        mout.errorOut("KeyboardInterrupt")
        exit()


def parsePDBAtomLine(
    line,
    res_index,
    atom_index,
    chain_counter,
    debug=False,
    alternative_site_warnings=True,
):
    from .atom import Atom
    import mout
    import string
    from .atom import Atom

    if debug:
        mout.debug("Attempting to parse atom with index: " + str(atom_index))
        print(line.strip())

    charge = None

    try:
        atom_name = line[12:16].strip()
        if debug:
            mout.var(f"{atom_index}.name", atom_name)

        residue = line[17:21].strip()
        if debug:
            mout.var(f"{atom_index}.residue", residue)

        try:
            pdb_index = int(line[6:12].strip())
        except:
            pdb_index = atom_index
        if debug:
            mout.var(f"{atom_index}.pdb_index", pdb_index)
        chain = line[21:22]
        if chain == " ":
            chain = string.ascii_uppercase[chain_counter % 26]
        if debug:
            mout.var(f"{atom_index}.chain", chain)
        res_number = line[22:26].strip()
        if debug:
            mout.var(f"{atom_index}.res_number", res_number)

        alt_site_str = line[16:17]
        if len(line[16:17].strip()) == 0:
            alt_site_str = None

        if alternative_site_warnings and alt_site_str:
            mout.warningOut(
                f"Alternative site in PDB! res={residue}, atom={atom_name}, res_number={res_number}"
            )
        if debug:
            mout.var(f"{atom_index}.alt_site_str", alt_site_str)

        position = []
        position.append(float(line[30:38].strip()))
        position.append(float(line[38:46].strip()))
        position.append(float(line[46:54].strip()))
        if debug:
            mout.var(f"{atom_index}.position", position)

        try:
            occupancy = float(line[55:61].strip())
        except:
            occupancy = None
        if debug:
            mout.var(f"{atom_index}.occupancy", occupancy)

        try:
            temp_factor = float(line[61:67].strip())
        except:
            temp_factor = None
        if debug:
            mout.var(f"{atom_index}.temp_factor", temp_factor)

        element = line[76:78].strip()
        if len(element) > 1:
            element = f"{element[0]}{element[1:].lower()}"
        if debug:
            mout.var(f"{atom_index}.element", f"###{element}###")

        chg_str = line[78:80].rstrip("\n")
        if len(chg_str.strip()) < 1:
            chg_str = None
        if debug:
            mout.var(f"{atom_index}.chg_str", chg_str)

        end = line[80:].strip()

        if line.startswith("HETATM"):
            hetatm = True
        else:
            hetatm = False
        if debug:
            mout.var(f"{atom_index}.hetatm", hetatm)

        if end:
            charge = float(line[80:89])

        atom = Atom(
            atom_name,
            pdb_index,
            pdb_index,
            position,
            residue,
            chain,
            res_number,
            occupancy=occupancy,
            temp_factor=temp_factor,
            heterogen=hetatm,
            charge_str=chg_str,
            alternative_site=alt_site_str,
            res_index=res_index,
            charge=charge,
            element=element,
        )

        if debug:
            atom.summary()

        if debug:
            raise Exception("Stopping DEBUG")

        return atom

    except Exception as e:
        mout.error(e)
        mout.error("Unsupported PDB line shown below:")
        mout.error(line)
        raise Exception("Unsupported PDB line shown above")


def parsePDBSSBOND(system, line):
    bond = [
        {
            "chain": line[15].strip(),
            "resname": line[11:14].strip(),
            "resid": int(line[17:21].strip()),
        },
        {
            "chain": line[29].strip(),
            "resname": line[25:28].strip(),
            "resid": int(line[31:35].strip()),
        },
        {
            "sym1": line[59:65].strip(),
            "sym2": line[66:72].strip(),
            "distance": float(line[73:78].strip()),
        },
    ]
    system._ssbonds.append(bond)


# def parseGRO(gro,systemName=None,fix_indices=True,fix_atomnames=True,verbosity=1,auto_ter=None):
def parseGRO(
    gro,
    systemName=None,
    fix_indices=True,
    fix_atomnames=True,
    autoname_chains=True,
    element_guess_warnings=True,
    verbosity=1,
    auto_ter=["DA3", "DT3", "DG3", "DC3"],
):
    """Parse a Gromacs structure"""
    import mcol
    import mout
    from .system import System
    from .chain import Chain

    assert gro.endswith(".gro")

    if verbosity > 0:
        mout.out(
            "parsing " + mcol.file + gro + mcol.clear + " ... ", end=""
        )  # user output

    import os

    residue = None
    chain = None
    last_residue_name = None
    last_residue_number = None
    last_chain_name = None
    res_counter = 0
    chain_counter = 0
    atom_counter = 1
    was_terminal = False

    if systemName is None:
        systemName = os.path.splitext(gro)[0]
    system = System(name=systemName)

    try:
        with open(gro, "r") as input_gro:
            first = True
            line_counter = 0
            # make_new_res = True
            for line in input_gro:
                line_counter += 1

                # parse the header
                if line_counter == 1:
                    sys_description = line.strip()
                elif line_counter == 2:
                    sys_atom_count = int(line.strip())

                # parse the rest
                else:

                    # check if last line
                    if len(line.strip()) < 40:
                        split_line = line.strip().split()
                        system.box = [
                            float(split_line[0]),
                            float(split_line[1]),
                            float(split_line[2]),
                        ]
                        break

                    # parse an atom line:
                    atom = parseGROAtomLine(
                        line,
                        res_counter,
                        atom_counter,
                        chain_counter,
                        element_guess_warning=element_guess_warnings,
                    )

                    # first atom
                    if line_counter == 3:
                        chain = Chain(atom.chain)
                        index = res_counter
                        residue = new_residue(
                            atom.residue, index, atom.res_number, atom.chain
                        )
                        # residue.addAtom(atom)
                        last_residue_name = atom.residue
                        last_residue_number = atom.res_number
                        last_chain_name = atom.chain

                    # check if a new residue is needed
                    make_new_res = False
                    if residue is None:
                        make_new_res = True
                    if last_residue_name != atom.residue:
                        make_new_res = True
                    if last_residue_number != atom.res_number:
                        make_new_res = True

                    # make the new residue
                    if make_new_res:
                        if residue is not None:
                            chain.add_residue(residue)
                            res_counter += 1
                        residue = new_residue(
                            atom.residue, index, atom.res_number, atom.chain
                        )

                    # add the atom to the residue
                    residue.addAtom(atom)
                    residue_type = residue.type

                    # check if a new chain is needed
                    make_new_chain = False
                    if residue is None:
                        # print("NONE-TER")
                        make_new_chain = True
                    if (
                        auto_ter is not None
                        and make_new_res
                        and last_residue_name in auto_ter
                    ):
                        make_new_chain = True
                        # print("AUTO-TER",residue.name)
                    if make_new_res and residue_type != last_residue_type:
                        make_new_chain = True
                        # print("TYPE-TER",residue.name)

                    # make the new chain
                    if make_new_res:
                        if make_new_chain:
                            if chain is not None:
                                system.add_chain(chain)
                                chain_counter += 1
                            chain = Chain(atom.chain)

                    # update variables
                    atom_counter += 1
                    last_residue_type = residue_type
                    last_residue_number = atom.res_number
                    last_residue_name = atom.residue

                    system.description = sys_description

    except FileNotFoundError:
        mout.errorOut("File " + mcol.file + gro + mcol.error + " not found", fatal=True)

    chain.add_residue(residue)
    system.add_chain(chain)

    if fix_indices:
        system.fix_indices()

    if fix_atomnames:
        system.fix_atomnames()

    if autoname_chains:
        system.autoname_chains()

    if verbosity > 0:
        mout.out("Done.")  # user output

    return system


def new_residue(name, index, number, chain):
    from .residue import res_type

    if res_type(name) == "PRO":
        if name in ["HIS", "HID", "HIE", "HSE", "HSD", "HSP"]:
            from .histidine import Histidine

            return Histidine(name, index, number, chain)
        else:
            from .amino import AminoAcid

            return AminoAcid(name, index, number, chain)
    elif res_type(name) == "DNA":
        from .nucleic import NucleicAcid

        return NucleicAcid(name, index, number, chain)
    elif name in ["ATP", "GTP", "CTP", "TTP", "OGTP"]:
        from .ntp import NucleobaseTriPhosphate

        return NucleobaseTriPhosphate(name, index, number, chain)
    else:
        from .residue import Residue

        return Residue(name, index, number, chain)


def parseGROAtomLine(
    line, res_index, atom_index, chain_counter, element_guess_warning=True
):
    import string
    from .atom import Atom

    res_number = line[0:5].strip()
    residue = line[5:10].strip()
    atom_name = line[11:15].strip()
    gro_index = line[15:21].strip()

    chain = string.ascii_uppercase[chain_counter % 26]

    position = []
    # position needs conversion nm --> Angstrom
    position.append(10.0 * float(line[21:29].strip()))
    position.append(10.0 * float(line[29:37].strip()))
    position.append(10.0 * float(line[37:45].strip()))

    velocity = []
    try:
        # velocity needs conversion: nm/ps --> Angstrom/fs
        velocity.append(0.01 * float(line[45:53].strip()))
        velocity.append(0.01 * float(line[53:61].strip()))
        velocity.append(0.01 * float(line[61:69].strip()))
    except:
        velocity = [0.0, 0.0, 0.0]

    hetatm = False

    atom = Atom(
        atom_name,
        atom_index,
        gro_index,
        position,
        residue,
        chain,
        res_number,
        velocity=velocity,
        res_index=res_index,
        element_guess_warning=element_guess_warning,
    )

    return atom


def writeCJSON(
    filename, system, use_atom_types=False, gulp_names=False, noPrime=False, verbosity=1
):
    if verbosity > 0:
        mout.out(
            "writing " + mcol.file + filename + mcol.clear + " ... ", end=""
        )  # user output

    # Check that the input is the correct class
    assert isinstance(system, System)

    # Load module
    import json

    if gulp_names:
        use_atom_types = True
        names = []
        temp_names = system.FF_atomtypes
        for name in temp_names:
            name = name[0] + "_" + name[1:]
            names.append(name)
    else:
        if not use_atom_types:
            names = system.atom_names(wRes=True, noPrime=noPrime)
        else:
            names = system.FF_atomtypes

    # Create the dictionary
    data = {}
    data["chemical json"] = 0
    data["name"] = system.name
    data["atoms"] = {
        "names": names,
        "elements": {"number": system.atomic_numbers},
        "coords": {
            "unit": "angstrom",
            "3d": [item for sublist in system.positions for item in sublist],
        },
        "charges": system.charges,
    }

    # Write the CJSON file
    with open(filename, "w") as f:
        json.dump(data, f, indent=4)

    if verbosity > 0:
        mout.out("Done.")  # user outpu


def writePDB(
    filename,
    system,
    verbosity=1,
    append=False,
    model=1,
    header=None,
    ssbonds=None,
    title=None,
    compound=None,
    write_model=True,
    charges=True,
    shift_name=0,
    hydrogen=True,
):
    import mcol
    import mout
    from .system import System
    from .group import AtomGroup

    if verbosity > 0:
        mout.out(
            f"writing {mcol.file}{filename}{mcol.clear} ... ", end=""
        )  # user output

    # Check that the input is the correct class
    assert isinstance(system, System) or isinstance(system, AtomGroup)

    end = "\n"

    if not append:

        if system._header_data is not None:

            strbuff = "".join(system._header_data)

        else:

            strbuff = ""

            if not header:
                strbuff += f"HEADER {filename}{end}"
            else:
                strbuff += f"HEADER {header}{end}"

            if not title:
                strbuff += f"TITLE  {system.name}{end}"
            else:
                strbuff += f"TITLE  {title}{end}"

            strbuff += "REMARK " + "generated by molparse.io.writePDB()" + end
            # strbuff += "REMARK "+end
            # strbuff += "REMARK "+"System Summary:"+end
            if isinstance(system, System):
                strbuff += "REMARK " + "# Chains:   " + str(system.num_chains) + end
                strbuff += "REMARK " + "# Residues: " + str(system.num_residues) + end
                if system.remarks:
                    for line in system.remarks:
                        strbuff += f"REMARK {line}" + end

            strbuff += "REMARK " + "# Atoms:    " + str(system.num_atoms) + end

        if ssbonds:
            if not system.ssbonds:
                mout.warningOut(
                    f"No SS bonds information available. Please run 'system.ssbond_guesser() first'."
                )
            else:
                for i, bond in enumerate(system.ssbonds):
                    line = constructPDBSSBONDLine(i, bond)
                    strbuff += line

    else:
        strbuff = ""

    if write_model:
        strbuff += "MODEL     " + str(model) + end

    for i, atom in enumerate(system.atoms):

        if not hydrogen and atom.symbol == "H":
            continue

        atomline = constructPDBAtomLine(atom, i, charges=charges, shift_name=shift_name)

        strbuff += atomline

        if atom.terminal:
            # print(atom, atom.number, 'TERMINAL2')
            strbuff += atom.ter_line

    if write_model:
        strbuff += "ENDMDL" + end

    if append:
        out_stream = open(filename, "a")
    else:
        out_stream = open(filename, "w")

    out_stream.write(strbuff)
    out_stream.close()

    if verbosity > 0:
        mout.out("Done.")  # user output


def modifyPDB(filename, atoms, copy_from=None):
    """in-place modification of a PDB file for improved performance when only modifying a certain subset of atoms

    filename: PDB to modify
    atoms: list of mp.Atom objects that will be modified in the PDB
    copy_from: (optional) copy {copy_from} --> {filename} before modifying

    """

    if copy_from:
        import shutil

        shutil.copyfile(copy_from, filename)

    import fileinput

    pdb_indices = [a.pdb_index for a in atoms]

    max_index = max(pdb_indices)
    min_index = min(pdb_indices)

    for line in fileinput.FileInput(filename, inplace=1):

        try:
            pdb_index = int(line[6:12].strip())

        except ValueError:
            print(line, end="")
            continue

        # check if the atom index matches a pdb_index in the atoms
        if (
            pdb_index >= min_index
            and pdb_index <= max_index
            and pdb_index in pdb_indices
        ):
            line = constructPDBAtomLine(atoms[pdb_indices.index(pdb_index)], pdb_index)

        print(line, end="")


def constructPDBAtomLine(atom, index, charges=True, shift_name=False, alt_sites=True):
    import mout
    import mcol

    end = "\n"

    strlist = []

    atom_serial = atom.number or atom.index
    residue_serial = atom.res_number

    if atom_serial is None:
        mout.warningOut(
            f"{mcol.varName}atom_serial{mcol.clear}{mcol.warning} is None",
            code=f"mp.io.constructPDBAtomLine({atom},{index})",
        )
        atom_serial = index
    if residue_serial is None:
        mout.errorOut(
            f"{mcol.varName}residue_serial{mcol.clear}{mcol.warning} is None",
            code=f"mp.io.constructPDBAtomLine({atom},{index})",
        )
        residue_serial = "    "

    if not atom.heterogen:
        strlist.append("ATOM  ")
    else:
        strlist.append("HETATM")

    atom_serial_str = str(atom_serial).rjust(5)
    if len(atom_serial_str) > 5:
        atom_serial_str = "XXXXX"

    residue_serial_str = str(residue_serial).rjust(4)
    if len(residue_serial_str) > 4:
        residue_serial_str = residue_serial_str[-4:]

    strlist.append(atom_serial_str)
    strlist.append(" ")
    if shift_name:
        strlist.append(f" {atom.name[:3].ljust(3)}")
    else:
        strlist.append(str(atom.name[:4]).ljust(4))
    if alt_sites and atom.alternative_site:
        strlist.append(str(atom.alternative_site))
    else:
        strlist.append(" ")
    # strlist.append(" ")
    if atom.residue is None:
        mout.errorOut(
            f"{mcol.varName}atom.residue{mcol.clear}{mcol.warning} is None",
            code=f"mp.io.constructPDBAtomLine({atom},{index})",
        )
        strlist.append("    ".ljust(4))
    else:
        res_str = str(atom.residue).ljust(4)
        if len(res_str) > 4:
            res_str = res_str[:4]
        strlist.append(res_str)
    if atom.chain is None:
        mout.errorOut(
            f"{mcol.varName}atom.chain{mcol.clear}{mcol.warning} is None",
            code=f"mp.io.constructPDBAtomLine({atom},{index})",
        )
        strlist.append(" ")
    else:
        strlist.append(str(atom.chain))
    strlist.append(residue_serial_str)
    strlist.append("    ")

    x_str = "{:.3f}".format(atom.x).rjust(8)
    y_str = "{:.3f}".format(atom.y).rjust(8)
    z_str = "{:.3f}".format(atom.z).rjust(8)

    strlist.append(x_str + y_str + z_str)

    if atom.occupancy is not None:
        strlist.append("{:.2f}".format(atom.occupancy).rjust(6))
    else:
        strlist.append("      ")
    if atom.temp_factor is not None:
        strlist.append("{:.2f}".format(atom.temp_factor).rjust(6))
    else:
        strlist.append("      ")

    strlist.append("          ")
    strlist.append(atom.species.rjust(2))

    if atom.charge_str is not None:
        strlist.append(atom.charge_str)
    # else:
    # strlist.append("  ")

    if charges and atom.charge is not None:
        strlist.append(f"{atom.charge:10.6f}")

    strlist.append(end)

    if atom.terminal:
        # print(atom, atom.number, 'TERMINAL1')
        atom_serial_str = str(atom_serial + 1).rjust(5)
        if len(atom_serial_str) > 5:
            atom_serial_str = "XXXXX"

        atom.ter_line = "TER   "
        atom.ter_line += atom_serial_str
        atom.ter_line += "      "
        atom.ter_line += atom.residue.ljust(4)
        atom.ter_line += atom.chain
        atom.ter_line += str(residue_serial).rjust(4)
        atom.ter_line += end

    return "".join(strlist)


def constructPDBSSBONDLine(n, bond):
    line = "SSBOND".rjust(6)
    line += str(n + 1).rjust(4)
    line += str(bond[0]["resname"]).rjust(4)
    line += str(bond[0]["chain"]).rjust(2)
    line += str(bond[0]["resid"]).rjust(5)
    line += str(bond[1]["resname"]).rjust(7)
    line += str(bond[1]["chain"]).rjust(2)
    line += str(bond[1]["resid"]).rjust(5)
    line += "                       ".rjust(23)
    line += str(bond[2]["sym1"]).rjust(7)
    line += str(bond[2]["sym2"]).rjust(7)
    line += "{:.2f}".format(bond[2]["distance"]).rjust(6)
    line += "  \n".rjust(2)

    return line


def writeGRO(filename, system, verbosity=1):
    import mcol
    import mout
    from .system import System

    if verbosity > 0:
        mout.out(
            "writing " + mcol.file + filename + mcol.clear + " ... ", end=""
        )  # user output

    # Check that the input is the correct class
    assert isinstance(system, System)

    end = "\n"

    strbuff = system.name + " (amp.io)" + end
    strbuff += str(system.num_atoms) + end

    atom_serial = 1
    residue_serial = 1

    for chain in system.chains:
        for residue in chain.residues:
            for atom in residue.atoms:

                residue_serial_str = str(atom.res_number).rjust(5)
                if len(residue_serial_str) > 5:
                    residue_serial_str = residue_serial_str[-5:]

                index = atom.pdb_index or atom.index
                atom_serial_str = str(index).rjust(4)
                if len(atom_serial_str) > 4:
                    atom_serial_str = atom_serial_str[-4:]

                strbuff += residue_serial_str
                strbuff += str(atom.residue).ljust(4)
                strbuff += " "
                strbuff += str(atom.name).rjust(5)
                strbuff += " "
                strbuff += atom_serial_str

                x_str = "{:.3f}".format(atom.x / 10.0).rjust(8)
                y_str = "{:.3f}".format(atom.y / 10.0).rjust(8)
                z_str = "{:.3f}".format(atom.z / 10.0).rjust(8)
                strbuff += x_str + y_str + z_str

                if atom.velocity is None:
                    x_str = "{:.4f}".format(0.0).rjust(8)
                    y_str = "{:.4f}".format(0.0).rjust(8)
                    z_str = "{:.4f}".format(0.0).rjust(8)
                else:
                    x_str = "{:.4f}".format(atom.velocity[0] * 100.0).rjust(8)
                    y_str = "{:.4f}".format(atom.velocity[1] * 100.0).rjust(8)
                    z_str = "{:.4f}".format(atom.velocity[2] * 100.0).rjust(8)
                strbuff += x_str + y_str + z_str

                strbuff += end

                atom_serial += 1

            residue_serial += 1

    if system.box is None:
        mout.warningOut("System has no box information. Using bbox")
        bbox = system.bbox
        x_str = "{:.5f}".format(bbox[0][1] - bbox[0][0]).rjust(10)
        y_str = "{:.5f}".format(bbox[1][1] - bbox[1][0]).rjust(10)
        z_str = "{:.5f}".format(bbox[2][1] - bbox[2][0]).rjust(10)
    else:
        x_str = "{:.5f}".format(system.box[0]).rjust(10)
        y_str = "{:.5f}".format(system.box[1]).rjust(10)
        z_str = "{:.5f}".format(system.box[2]).rjust(10)
    strbuff += x_str + y_str + z_str
    strbuff += end

    out_stream = open(filename, "w")

    out_stream.write(strbuff)
    out_stream.close()

    if verbosity > 0:
        mout.out("Done.")  # user output


def parseXYZ(xyz, index=":", verbosity=1):
    import mout
    import mcol

    assert xyz.endswith(".xyz")

    try:
        index = int(index)
    except:
        if index == ":":
            if verbosity > 0:
                mout.warningOut("Parsing all models in " + mcol.file + xyz)
        else:
            mout.errorOut("Unsupported index: '" + str(index) + "'", fatal=True)

    if index == ":":

        all_models = []

        import subprocess

        num_lines = int(subprocess.check_output(f"cat {xyz} | wc -l", shell=True))

        with open(xyz, "r") as input_xyz:
            for line in input_xyz:
                num_atoms = int(line)
                break

        if (num_lines / (num_atoms + 2)) != (num_lines // (num_atoms + 2)):
            mout.errorOut(
                "Wrong number of lines, check XYZ formatting is correct.", fatal=True
            )

        num_models = num_lines // (num_atoms + 2)

        for i in range(num_models):
            all_models.append(parseXYZ(xyz, index=i, verbosity=verbosity - 1))

        return all_models

    elif index < 0:

        import subprocess

        num_lines = int(subprocess.check_output(f"cat {xyz} | wc -l", shell=True))

        with open(xyz, "r") as input_xyz:
            for line in input_xyz:
                num_atoms = int(line)
                break

        if (num_lines / (num_atoms + 2)) != (num_lines // (num_atoms + 2)):
            mout.errorOut(
                "Wrong number of lines, check XYZ formatting is correct.", fatal=True
            )

        num_models = num_lines // (num_atoms + 2)

        return parseXYZ(xyz, index=num_models + index - 1, verbosity=verbosity)

    else:

        from .group import AtomGroup

        with open(xyz, "r") as input_xyz:

            start_line = 0

            for j, line in enumerate(input_xyz):

                # print(line)

                if j == 0:
                    num_atoms = int(line)
                    start_line = index * (num_atoms + 2)
                    continue

                elif j == start_line + 1:

                    try:
                        # get the model information
                        i_str, E_str = line.strip().split(",")
                        i = int(i_str.split("=")[-1])
                        E = float(E_str.split("=")[-1])
                        system = AtomGroup(f"{xyz}, image={i}")
                        system._energy = E
                        system._traj_index = i
                    except ValueError:
                        mout.warningOut(
                            "XYZ image header string has an unsupported format"
                        )
                        system = AtomGroup(f"{xyz}, {line.strip()}")

                    # create the system object
                    atom_counter = 0
                    continue

                elif j < start_line + 1:
                    continue

                if num_atoms == atom_counter:
                    break

                atom = parseXYZAtomLine(line, atom_counter)
                system.add_atom(atom)
                atom_counter += 1

            if j < start_line:
                mout.errorOut(f"XYZ has no image with index {index}")

        return system


def parseXYZAtomLine(line, atom_index):
    from .atom import Atom

    split_line = line.strip().split()

    if len(split_line) == 4:
        s, x, y, z = split_line
    else:
        s, x, y, z = split_line[:4]

    x = float(x)
    y = float(y)
    z = float(z)

    atom = Atom(s, index=atom_index, position=[x, y, z])

    return atom


def modifyCIF(
    cif_file,
    out_file,
    remove_atomtypes: "list[str] | None" = None,
):

    from gemmi import cif

    # from rdkit import Chem, Geometry
    from .rdkit.xca_utils import strip_quotes, BOND_TYPES

    if remove_atomtypes:
        remove_atomtypes = [t.upper() for t in remove_atomtypes]

    # mol = Chem.RWMol()
    # conf = Chem.Conformer()

    doc = cif.read(str(cif_file))

    # Diamond CIFs have two blocks, but the one we want will be named data_comp_LIG
    block = doc.find_block("comp_LIG")

    # Other CIFs have unpredictable block names, so let's hope there is only one
    if not block:
        block = doc.sole_block()
    if not block:
        print("sole block not found")
        return None

    atom_symbols = block.find_loop("_chem_comp_atom.type_symbol")

    # find offending atoms
    atoms = {}
    ligand_name = None
    del_list = []
    for i, s in enumerate(atom_symbols):
        if s in remove_atomtypes:
            del_list.append(i)

    # clean up atom list
    if del_list:

        loop = atom_symbols.get_loop()

        new_values = []
        for tag in loop.tags:

            col = block.find_loop(tag)
            l = list(col)

            if "ordinal" in tag:
                l = [str(i + 1) for i in range(len(l) - len(del_list))]

            else:
                for i in reversed(del_list):
                    del l[i]

            new_values.append(l)

        loop.set_all_values(new_values)

    atom1 = block.find_loop("_chem_comp_bond.atom_id_1")
    atom2 = block.find_loop("_chem_comp_bond.atom_id_2")

    # find offending bonds
    del_list = []
    for i, (a1, a2) in enumerate(zip(atom1, atom2)):
        if a1 in remove_atomtypes or a2 in remove_atomtypes:
            del_list.append(i)

    # clean up bond list
    if del_list:

        loop = atom1.get_loop()

        new_values = []
        for tag in loop.tags:

            col = block.find_loop(tag)
            l = list(col)

            if "ordinal" in tag:
                l = [str(i + 1) for i in range(len(l) - len(del_list))]

            else:
                for i in reversed(del_list):
                    del l[i]

            new_values.append(l)

        loop.set_all_values(new_values)

    # write the modified file
    doc.write_file(out_file)


def parseCIF(
    cif_file,
    verbosity=1,
    split_chains_by_type: bool = True,
    debug: bool = False,
    **kwargs,
):
    """Parse a PDBx/mmCIF structure"""

    from gemmi import cif
    from rich import print
    import mrich
    from pandas import DataFrame
    from .system import System
    from .atom import Atom
    from .chain import Chain

    if kwargs and verbosity > 0:
        mrich.warning("Unused keyword arguments", kwargs)

    doc = cif.read(str(cif_file))

    greeted = set()

    block = doc.sole_block()  # mmCIF has exactly one block

    ### initialise the system

    sys = System(block.name)

    ### CRYST1 INFO

    cell_dict = block.get_mmcif_category("_cell.")
    symmetry_dict = block.get_mmcif_category("_symmetry.")

    if debug:
        mrich.print("cell_dict:", cell_dict)
        mrich.print("symmetry_dict:", symmetry_dict)

    sys.add_CRYST1(
        a=cell_dict["length_a"][0],
        b=cell_dict["length_b"][0],
        c=cell_dict["length_c"][0],
        alpha=cell_dict["angle_alpha"][0],
        beta=cell_dict["angle_beta"][0],
        gamma=cell_dict["angle_gamma"][0],
        space_group=symmetry_dict["space_group_name_H-M"][0],
        z=cell_dict["Z_PDB"][0],
        debug=debug,
    )

    ### ATOM records

    atom_dict = block.get_mmcif_category("_atom_site.")

    atom_df = DataFrame(atom_dict)

    if verbosity > 0:
        gen = mrich.track(
            atom_df.iterrows(), prefix="Parsing atom entries", total=len(atom_df)
        )
    else:
        gen = atom_df.iterrows()

    chain = None
    residue = None
    make_new_chain = True
    make_new_residue = True
    res_counter = 0

    for i, row in gen:

        atom_name = str(row.label_atom_id)
        atom_number = int(row.id)
        atom_element = str(row.type_symbol)
        atom_charge = row.pdbx_formal_charge
        if atom_charge:
            atom_charge = int(atom_charge)
        atom_position = [float(row.Cartn_x), float(row.Cartn_y), float(row.Cartn_z)]
        atom_occupancy = float(row.occupancy)
        atom_heterogen = row.group_PDB == "HETATM"

        residue_name = str(row.label_comp_id)
        residue_number = int(row.label_seq_id)

        chain_name = str(row.auth_asym_id)

        alternative_site = row.label_alt_id
        if alternative_site:
            alternative_site = str(alternative_site)
        else:
            alternative_site = None

        # some minor validation
        model_num = int(row.pdbx_PDB_model_num)
        assert model_num == 1

        if debug:
            mrich.debug(
                row.group_PDB,
                atom_name,
                i,
                atom_number,
                atom_position,
                residue_name,
                chain_name,
                residue_number,
            )

        atom = Atom(
            name=atom_name,
            index=i,
            pdb_index=atom_number,
            position=atom_position,
            residue=residue_name,
            chain=chain_name,
            res_number=residue_number,
            charge=atom_charge,
            # FF_atomtype=None,
            # mass=None,
            # LJ_sigma=None,
            # LJ_epsilon=None,
            occupancy=atom_occupancy,
            # temp_factor=None,
            heterogen=atom_heterogen,
            # charge_str=None,
            # velocity=None,
            alternative_site=alternative_site,
            res_index=None,  ### DOES THIS NEED TO BE SET?
            element=atom_element,
        )

        if chain:
            make_new_chain = chain.name != chain_name
            if split_chains_by_type and residue:
                make_new_chain = make_new_chain or chain.type != residue.type

        if residue:
            make_new_residue = (
                residue.name != residue_name or residue.number != residue_number
            )

        if not chain or make_new_chain:

            if chain:
                sys.add_chain(chain)

            chain = Chain(chain_name)

        if not residue or make_new_residue:

            if residue:
                chain.add_residue(residue)
                res_counter += 1

            residue = new_residue(
                residue_name,
                res_counter,
                residue_number,
                chain_name,
            )

        residue.addAtom(atom)

    chain.add_residue(residue)
    sys.add_chain(chain)

    sys.fix_indices()

    return sys



from .group import AtomGroup

import mrich as logger


class Chain(AtomGroup):
    """Class containing covalently bonded Residue objects

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(self, name):

        super(Chain, self).__init__(name)

        from .list import NamedList

        self.residues = NamedList()

        self.index = None
        self.fix_names()

        self._parent = None

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, obj):
        # import weakref
        # self._parent = weakref.ref(obj)
        self._parent = obj

    def fix_names(self):
        """Ensure child-classes have correct parent name"""
        for residue in self.residues:
            residue.chain = self._name
        for atom in self.atoms:
            atom.chain = self._name

    @property
    def num_residues(self):
        """Number of child residues (int)"""
        return len(self.residues)

    def add_residue(self, residue):
        """Add a Residue to the chain"""
        import mcol
        import mout
        from .residue import Residue

        assert isinstance(residue, Residue)
        residue.set_chain_number(self.index)
        residue.set_chain_char(self.name)
        residue.parent = self

        # remove any termini
        if self.residues:
            for atom in self.residues[-1].atoms:
                atom.terminal = None
                atom.ter_line = None

        self.residues.append(residue)
        if self.num_residues > 1 and self.residues[-1].type != self.type:
            mout.errorOut(
                f"Differing residue types in same chain {self.residues[-1]} ({self.residues[-1].type}), {self.residues[0]} ({self.type})"
            )

    @property
    def type(self):
        """Classification of Residues in this Chain"""
        return self.residues[0].type

    @property
    def res_names(self):
        """All child Residue names (list)"""
        return [res.name for res in self.residues]

    def atom_names(self, wRes=False, noPrime=False):
        """All child Atom names (list)"""
        names_list = []
        for residue in self.residues:
            names_list += residue.atom_names(wRes=wRes, noPrime=noPrime)
        return names_list

    @property
    def atoms(self):
        """All child Atom objects (list)"""
        atoms = []
        for residue in self.residues:
            atoms += residue.atoms
        return atoms

    @property
    def FF_atomtypes(self):
        """All child Atom atomtypes (list)"""
        atomtype_list = []
        for residue in self.residues:
            atomtype_list += residue.FF_atomtypes
        return atomtype_list

    def index_from_name(self, namestring):
        """Find the index of an atom by its name.
        Assumes formatting resname_atomname"""
        import mcol
        import mout

        search_residue, search_atom = namestring.split("_")

        for index, atom in enumerate(self.atoms):
            if atom.residue == search_residue and atom.name == search_atom:
                return index

        mout.errorOut(
            "Atom "
            + mcol.arg
            + search_atom
            + mcol.error
            + " could not be found in residue"
            + mcol.arg
            + search_residue
            + " of chain "
            + mcol.arg
            + self.name,
            fatal=True,
            code="Chain.1",
        )

    def copy(self, fast=False):
        """return a deepcopy of the Chain"""
        if fast:
            new_chain = Chain(self.name)
            for residue in self.residues:
                new_chain.add_residue(residue.copy(fast=fast))
            return new_chain
        else:
            import copy

            return copy.deepcopy(self)

    def add_atom(self, atom):
        """add an atom to the chain"""
        from .residue import Residue

        if self.residues and atom.is_in_residue(self.residues[-1]):
            self.residues[-1].add_atom(atom)
        else:

            residue = Residue(atom.residue, number=atom.res_number)
            residue.index = atom.res_index
            residue.chain = atom.chain
            residue.add_atom(atom)
            self.add_residue(residue)

    @property
    def children(self):
        return self.residues

    def remove_residues(
        self,
        *,
        names: "list | None" = None,
        indices: "list | None" = None,
        numbers: "list | None" = None,
        verbosity: int = 2,
    ) -> int:

        assert names or indices or numbers, "must supply names, indices or numbers"

        import mcol

        mout = logger

        number_deleted = 0

        if indices:
            del_list = indices
            f = lambda x: x.index
        elif numbers:
            del_list = numbers
            f = lambda x: x.number
        else:
            del_list = names
            f = lambda x: x.name

        for index, residue in reversed(list(enumerate(self.residues))):
            if f(residue) in del_list:
                del self.residues[index]
                number_deleted += 1
                if verbosity > 1:
                    mout.warning(f"Removed residue {residue.name_number_chain_str}")

        if verbosity > 0:
            mout.var("#residues deleted", number_deleted)

        return number_deleted

    @property
    def sequence(self) -> str:
        """Amino acid sequence"""
        assert self.type == "PRO"
        return "".join([r.letter for r in self.residues])

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):

        assert isinstance(name, str)

        self._name = name
        for residue in self.residues:
            residue.chain = name
            for atom in residue.atoms:
                atom.chain = name



standard = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.9,
    "rotation": "",
    "celllinewidth": 0.0,
}

standard_rot = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.9,
    "rotation": "-90x,65z",
    "celllinewidth": 0.0,
}

standard_bonds = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.5,
    "celllinewidth": 0.0,
}

standard_cell = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": True,
    "radii": 0.9,
    "celllinewidth": 0.07,
}

gif_standard = {"loop": 0, "fps": 25, "subrectangles": True}



def methylate(residue, site, position=None):
    import numpy as np
    from ..atom import Atom

    if position is None:
        position = site.np_pos + np.array([1.538, 0, 0])

    l0_CB_HB = 1.111  # Angstrom
    a0_HB_CB_CS = 110.10  # degrees

    # place CB
    CB = Atom(
        "CB",
        position=position,
        residue=residue.name,
        chain=residue.chain,
        res_number=residue.number,
    )
    residue.add_atom(CB)

    # distances
    dist_Z = np.sin(np.pi / 180 * (a0_HB_CB_CS - 90)) * l0_CB_HB
    r_XY = np.cos(np.pi / 180 * (a0_HB_CB_CS - 90)) * l0_CB_HB

    # vectors
    vec_CA_CB = CB - site
    vec_X = np.cross(vec_CA_CB, [1, 0, 0])
    vec_X /= np.linalg.norm(vec_X)
    vec_Y = np.cross(vec_CA_CB, vec_X)
    vec_Y /= np.linalg.norm(vec_Y)

    # place HB1
    dist_X = r_XY * np.sin(0.0)
    dist_Y = r_XY * np.cos(0.0)
    pos_HB1 = CB + vec_CA_CB * dist_Z + vec_X * dist_X + dist_Y * vec_Y
    HB1 = Atom(
        "HB1",
        position=pos_HB1,
        residue=residue.name,
        chain=residue.chain,
        res_number=residue.number,
    )
    residue.add_atom(HB1)

    # place HB2
    dist_X = r_XY * np.sin(2 / 3 * np.pi)
    dist_Y = r_XY * np.cos(2 / 3 * np.pi)
    pos_HB2 = CB + vec_CA_CB * dist_Z + vec_X * dist_X + dist_Y * vec_Y
    HB2 = Atom(
        "HB2",
        position=pos_HB2,
        residue=residue.name,
        chain=residue.chain,
        res_number=residue.number,
    )
    residue.add_atom(HB2)

    # place HB3
    dist_X = r_XY * np.sin(4 / 3 * np.pi)
    dist_Y = r_XY * np.cos(4 / 3 * np.pi)
    pos_HB3 = CB + vec_CA_CB * dist_Z + vec_X * dist_X + dist_Y * vec_Y
    HB3 = Atom(
        "HB3",
        position=pos_HB3,
        residue=residue.name,
        chain=residue.chain,
        res_number=residue.number,
    )
    residue.add_atom(HB3)



from .residue import Residue, res_type
from .group import AtomGroup


def alphabet(name):
    if name == "DA":
        return "A"
    elif name == "DT":
        return "T"
    elif name == "DG":
        return "G"
    elif name == "DC":
        return "C"
    elif name == "DA3":
        return "A"
    elif name == "DT3":
        return "T"
    elif name == "DG3":
        return "G"
    elif name == "DC3":
        return "C"
    elif name == "DA5":
        return "A"
    elif name == "DT5":
        return "T"
    elif name == "DG5":
        return "G"
    elif name == "DC5":
        return "C"
    else:
        import mout

        mout.errorOut(f"Unsupported nucleic acid residue name: {name}")
        return None


def longname(name):
    if name == "DA":
        return "Adenine"
    elif name == "DT":
        return "Thymine"
    elif name == "DG":
        return "Guanine"
    elif name == "DC":
        return "Cytosine"
    elif name == "DA3":
        return "Adenine 3-Terminus"
    elif name == "DT3":
        return "Thymine 3-Terminus"
    elif name == "DG3":
        return "Guanine 3-Terminus"
    elif name == "DC3":
        return "Cytosine 3-Terminus"
    elif name == "DA5":
        return "Adenine 5-Terminus"
    elif name == "DT5":
        return "Thymine 5-Terminus"
    elif name == "DG5":
        return "Guanine 5-Terminus"
    elif name == "DC5":
        return "Cytosine 5-Terminus"
    else:
        import mout

        mout.errorOut(f"Unsupported nucleic acid residue name: {name}")
        return None


class NucleicAcid(Residue):
    """Class for Nucleic Acid Residue

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(
        self,
        name: str,
        index: int = None,
        number: int = None,
        chain: str = None,
        atoms=None,
    ):
        assert res_type(name) == "DNA"
        super(NucleicAcid, self).__init__(name, index, number, chain, atoms)

        self._backbone = None
        self._nucleobase = None

    @property
    def is_purine(self):
        return self.name in ["DG", "DA"]

    @property
    def is_pyrimidine(self):
        return self.name in ["DC", "DT"]

    @property
    def letter(self):
        """Nucleic acid alphabet"""
        return alphabet(self.name)

    @property
    def longname(self):
        """Nucleic acid chemical name"""
        return longname(self.name)

    @property
    def type(self):
        """Fixed typing"""
        return "DNA"

    @property
    def backbone(self):
        """Get backbone atoms"""
        if not self._backbone:
            self._backbone = self.get_atom(self.bb_names)
        return self._backbone

    def remove_backbone(self, add_link=True):
        """Remove backbone atoms"""

        if add_link:
            self.get_atom("C1'").set_name("HLNK")

        for atom in self.backbone:
            if not atom:
                continue
            self.delete_atom(atom.name)

    @property
    def nucleobase(self):
        """Get nucleobase atoms"""
        if not self._nucleobase:
            self._nucleobase = self.get_atom(self.nonbb_names)
        return self._nucleobase

    @property
    def bb_names(self):
        return [
            "H5T",
            "O5'",
            "C5'",
            "H5'",
            "H5''",
            "C4'",
            "H4'",
            "O4'",
            "C1'",
            "H1'",
            "C2'",
            "H2'",
            "H2''",
            "C3'",
            "H3'",
            "O3'",
            "P",
            "O1P",
            "O2P",
            "H3T",
        ]

    @property
    def nonbb_names(self):
        """Get nucleobase atoms"""
        return [n for n in self.atom_names() if n not in self.bb_names]

    def make_5ter(residue):
        """Create a 5-Terminus"""

        # Append 5 to residue name
        if not residue.name.endswith("5"):
            residue.name += "5"

        # Remove HTER/H5T
        residue.delete_atom("HTER")
        residue.delete_atom("H5T")
        residue.delete_atom("HO5'")
        residue.delete_atom("OXT")
        residue.delete_atom("O5T")
        residue.delete_atom("O1P")
        residue.delete_atom("O2P")
        residue.delete_atom("OP1")
        residue.delete_atom("OP2")

        # Rename P->H5T
        atom = residue.get_atom("P")
        if atom is not None:
            atom.name = "H5T"

    def make_3ter(residue):
        """Create a 3-Terminus"""

        # Append 3 to residue name
        if not residue.name.endswith("3"):
            residue.name += "3"

        # Remove HTER/H3T
        residue.delete_atom("O1P3")
        residue.delete_atom("O2P3")
        residue.delete_atom("O3T")
        residue.delete_atom("H3T")
        residue.delete_atom("HO3'")
        residue.delete_atom("HCAP")

        # Rename P3->H3T
        atom = residue.get_atom("P3")
        if atom is not None:
            atom.name = "H3T"

    def flip(self):

        import mout
        import mcol

        mout.header(
            f"Flipping {mcol.arg}{self.longname} {mcol.clear}{mcol.bold}({mcol.arg}{self.name_number_str}{mcol.clear}{mcol.bold})"
        )

        nucleobase_group = AtomGroup.from_any(
            f"{self.name}.nucleobase", self.nucleobase
        )

        if self.is_purine:
            source = nucleobase_group.atoms["N9"][0]
            target = nucleobase_group.atoms["O6"][0]

        vector = target - source

        nucleobase_group.rotate(180, vector, source.np_pos)

        for a1, a2 in zip(self.nucleobase, nucleobase_group.atoms):
            a1.position = a2.position

    def mutate(self, newname, show=False):
        """Mutate this nucleic acid to another"""

        import mout
        import mcol

        mout.headerOut(
            f"Mutating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg}{longname(newname)}{mcol.clear + mcol.bold} ({mcol.result}{self.letter}{self.number}{alphabet(newname)}{mcol.clear + mcol.bold})"
        )

        if self.name == newname:
            mout.warningOut("Skipping mutation with self == target!")
            return

        import os

        amp_path = os.path.dirname(__file__)

        from .io import parse

        if newname in ["DA", "DT"]:
            ref_sys = parse(f"{amp_path}/ref/AT_FLAT.pdb", verbosity=0)
        elif newname in ["DG", "DC"]:
            ref_sys = parse(f"{amp_path}/ref/GC_FLAT.pdb", verbosity=0)
        else:
            mout.errorOut("Unsupported target mutation")

        rem = [c for c in ref_sys.chain_names if c not in newname[-1]][0]
        ref_sys.remove_chain(rem, verbosity=0)
        ref_sys.fix_indices()
        ref_res = ref_sys.residues[0]

        if self.is_purine and ref_res.is_purine:

            # get atoms to align by
            names = ["N9", "C6", "C2"]
            start = ref_res.get_atom(names)
            target = self.get_atom(names)

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # delete intersecting atoms
            ref_res.delete_atom("H9", verbosity=0)

        elif self.is_pyrimidine and ref_res.is_pyrimidine:

            # get atoms to align by
            names = ["C6", "C4", "C2"]
            start = ref_res.get_atom(names)
            target = self.get_atom(names)

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N1") - ref_res.get_atom("N1"))

            # delete intersecting atoms
            ref_res.delete_atom("H1", verbosity=0)

        elif self.is_purine and ref_res.is_pyrimidine:
            mout.warningOut("Purine --> Pyrimidine")

            # get atoms to align by
            start = ref_res.get_atom(["C6", "C4", "C2"])
            target = self.get_atom(["C4", "C6", "C2"])

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N9") - ref_res.get_atom("N1"))

            # delete intersecting atoms
            ref_res.delete_atom("H1", verbosity=0)

        elif self.is_pyrimidine and ref_res.is_purine:
            mout.warningOut("Pyrimidine --> Purine")

            # get atoms to align by
            start = ref_res.get_atom(["C4", "C6", "C2"])
            target = self.get_atom(["C6", "C4", "C2"])

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N1") - ref_res.get_atom("N9"))

            # delete intersecting atoms
            ref_res.delete_atom("H9", verbosity=0)

        self.name = newname

        if show:
            view_sys = ref_sys.copy()
            for atom in self.atoms:
                view_sys.add_atom(atom)
            view_sys.plot3d(extra, 1.0)

        # delete intersecting atoms
        self.delete_atom(self.nonbb_names, verbosity=0)

        # add the reference nucleobase
        for atom in ref_res.atoms:
            atom._NUMBER = None
            self.add_atom(atom)

        if self.letter == "G":
            from .guanine import Guanine

            self = Guanine.from_nucleic(self)

        return self

    def oxidise(self):

        assert self.name == "DG"

        import numpy as np
        from .atom import Atom

        import mout
        import mcol

        mout.header(
            f"Oxidising {mcol.arg}{self.longname}{mcol.header} --> {mcol.result}{mcol.bold}8-Oxo-7,8-dihydroguanine (DOG)"
        )

        # H8 --> O8
        self.get_atom("H8").name = "O8"

        # +H7
        N7 = self.get_atom("N7")
        N9 = self.get_atom("N9")
        C4 = self.get_atom("C4")
        vec = N7 - (N9 + C4) / 2
        vec /= np.linalg.norm(vec)
        atom = Atom("H7")
        atom.position = N7 + vec
        self.add_atom(atom)

        # +H2
        N2 = self.get_atom("N2")
        C2 = self.get_atom("C2")
        N3 = self.get_atom("N3")
        vec = np.cross(N2 - C2, N3 - C2)
        vec /= np.linalg.norm(vec)
        atom = Atom("H2")
        atom.position = C2 + vec
        self.add_atom(atom)

        # +H3
        C6 = self.get_atom("C6")
        vec = N3 - C6
        vec /= np.linalg.norm(vec)
        atom = Atom("H3")
        atom.position = N3 + vec
        self.add_atom(atom)

        self.name = "DOG"



def to_alanine(residue):
    """Remove the amino acid's sidechain and replace with a methyl group"""
    from ..amino import AminoAcid

    assert isinstance(residue, AminoAcid)

    import mout
    from .functional import methylate

    # get the original amino-acids CB position
    pos_CB = residue.CB.np_pos

    # delete sidechain
    names = [a.name for a in residue.sidechain]
    residue.delete_atom(names, verbosity=1)

    # add the methyl group
    methylate(residue, residue.CA, pos_CB)

    # rename the residue
    residue.name = "ALA"


def replace_sidechain(residue, reference):
    """Remove the amino acid's sidechain and replace with one from the reference residue"""

    import mout
    from ..amino import AminoAcid

    assert isinstance(residue, AminoAcid)
    assert isinstance(reference, AminoAcid)

    # common sidechain atoms
    common = list(set(residue.sidechain_names).intersection(reference.sidechain_names))
    nonH_common = [n for n in common if not n.startswith("H")]
    mout.out(
        f"{residue} and {reference} share {len(nonH_common)} non-hydrogen side chain atoms"
    )

    if len(nonH_common) > 1:

        # align using these atoms
        align = ["CA"] + nonH_common[:3]

        # align the reference to the target
        res_map = [residue.get_atom(n) for n in align]
        ref_map = [reference.get_atom(n) for n in align]
        reference.align_by_posmap([ref_map, res_map])

        # replace the sidechain
        residue.remove_sidechain(verbosity=0)
        for atom in reference.sidechain:
            residue.add_atom(atom)

        # rename the residue
        residue.name = reference.name

    else:

        mout.var("nonH_common", nonH_common)

        mout.errorOut("Unsupported (residues do not share enough sidechain atoms)")
        raise NotImplementedError(
            "Unsupported (residues do not share enough sidechain atoms)"
        )



from .amino import AminoAcid


class Histidine(AminoAcid):
    """Class for Histidine Amino Acid Residue

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(self, name, index, number, chain):
        super(Histidine, self).__init__(name, index, number, chain)

        assert len(name) == 3
        assert name in ["HIS", "HID", "HIE", "HSE", "HSD", "HSP"]

    @property
    def ND(self):
        """Get Nitrogen-Delta Atom"""
        return self.get_atom("ND1", verbosity=False)

    @property
    def CG(self):
        """Get Carbon-Gamma Atom"""
        return self.get_atom("CG", verbosity=False)

    @property
    def CD(self):
        """Get Carbon-Delta Atom"""
        return self.get_atom("CD2", verbosity=False)

    @property
    def NE(self):
        """Get Nitrogen-Epsilon Atom"""
        return self.get_atom("NE2", verbosity=False)

    @property
    def HD(self):
        """Get Hydrogen-Delta Atom"""
        return self.get_atom("HD1", verbosity=False)

    @property
    def HE(self):
        """Get Hydrogen-Epsilon Atom"""
        return self.get_atom("HE2", verbosity=False)

    @property
    def has_HD(self):
        """Get Hydrogen-Delta Atom"""
        return bool(self.HD)

    @property
    def has_HE(self):
        """Get Hydrogen-Epsilon Atom"""
        return bool(self.HE)

    def protonate(self, delta=True, epsilon=True, verbosity=1):
        """Mutate this histidine residue to another protonation state"""

        import mout
        import mcol
        import numpy as np
        from .atom import Atom

        if verbosity > 0:
            if delta and epsilon:
                mout.headerOut(
                    f"Protonating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg} Histidine (Protonated)"
                )
            elif delta and not epsilon:
                mout.headerOut(
                    f"Protonating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg} Histidine (Delta)"
                )
            elif epsilon and not delta:
                mout.headerOut(
                    f"Protonating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg} Histidine (Epsilon)"
                )
            else:
                mout.errorOut(f"Unsupported protonation {delta=} {epsilon=}")
                return

        # warning
        if not delta and not epsilon:
            mout.warningOut("Neither site protonated!")

        # add delta-hydrogen
        if delta and not self.has_HD:

            if verbosity > 1:
                mout.warningOut("Protonating delta site")

            length = 1.00  # Ang

            site = self.ND
            opposite = (self.NE + self.CD) / 2
            vec = site.np_pos - opposite
            vec = length * vec / np.linalg.norm(vec)
            pos = site + vec

            if not epsilon and self.has_HE:
                atom = self.HE
                atom.name = "HD1"
            else:
                atom = Atom("HD1")
                self.add_atom(atom)
            atom.position = pos

        # add epsilon-hydrogen
        if epsilon and not self.has_HE:

            if verbosity > 1:
                mout.warningOut("Protonating epsilon site")

            length = 1.00  # Ang

            site = self.NE
            opposite = (self.ND + self.CG) / 2
            vec = site.np_pos - opposite
            vec = length * vec / np.linalg.norm(vec)
            pos = site + vec

            if not delta and self.has_HD:
                atom = self.HD
                atom.name = "HE2"
            else:
                atom = Atom("HE2")
                self.add_atom(atom)
            atom.position = pos

        # remove delta-hydrogen
        if not delta and self.has_HD:
            self.delete_atom("HD1")

        # remove epsilon-hydrogen
        if not epsilon and self.has_HE:
            self.delete_atom("HE2")

        # rename the residue
        if self.has_HD and not self.has_HE:
            self.name = "HSD"
        elif self.has_HE and not self.has_HD:
            self.name = "HSE"
        elif self.has_HD and self.has_HE:
            self.name = "HSP"



# to-do:

"""

improve performance by creating a smaller subsystem just of varying atoms, interpolating those and writing them to disk. Then use SED or other bash tools to insert the modifications into the original pdb

"""


def interpolate(
    input1,
    input2,
    frames=10,
    verbosity=2,
    smooth=False,
    indices=None,
    frame_padding=0,
    grid=False,
    ratios=None,
):
    import mout
    from ase import Atoms as aseatoms
    from .system import System

    if verbosity > 0:
        mout.headerOut("Interpolating...")
    if verbosity > 1:
        mout.varOut("input1", str(input1))
        mout.varOut("input2", str(input2))
        mout.varOut("frames", frames)
        mout.varOut("frame_padding", frame_padding)

    if isinstance(input1, aseatoms):
        mout.out("Input1 = ase.Atoms")
        mout.errorOut("Not supported yet", fatal=True)
    elif isinstance(input1, System):
        mout.out("Input1 = amp.system")
    else:
        mout.out(str(type(input1)))
        mout.errorOut("Not supported yet", fatal=True)

    if isinstance(input2, aseatoms):
        mout.out("Input2 = ase.Atoms")
        mout.errorOut("Not supported yet", fatal=True)
    elif isinstance(input2, System):
        mout.out("Input2 = amp.system")
    else:
        mout.out(str(type(input2)))
        mout.errorOut("Not supported yet", fatal=True)

    # these copy calls take a very long time
    system = input1.copy(alt=False)

    if not grid:

        system_array = []
        iterange = range(-frame_padding, frames + frame_padding)

        print(iterange)

        if indices is None and ratios is not None:
            mout.warningOut("Ratios not used as no indices passed!")
        elif indices is not None:
            assert len(ratios) == len(indices)

        # mout.progress(0,frames+2*frame_padding)

        for i in iterange:

            system.name = "Interpolation Frame " + str(i)

            if indices is not None:
                for j, index in enumerate(indices):
                    atom = system.atoms[index]

                    if ratios is not None:
                        atom.position = custom_interpolate(
                            input1.atoms[index].np_pos,
                            input2.atoms[index].np_pos,
                            ratios[j][i],
                        )
                    elif smooth:
                        atom.position = smooth_interpolate(
                            input1.atoms[index].np_pos,
                            input2.atoms[index].np_pos,
                            frames,
                            i,
                        )
                    else:
                        atom.position = simple_interpolate(
                            input1.atoms[index].np_pos,
                            input2.atoms[index].np_pos,
                            frames,
                            i,
                        )

            else:

                for index, atom in enumerate(system.atoms):
                    if smooth:
                        atom.position = smooth_interpolate(
                            input1.atoms[index].np_pos,
                            input2.atoms[index].np_pos,
                            frames,
                            i,
                        )
                    else:
                        atom.position = simple_interpolate(
                            input1.atoms[index].np_pos,
                            input2.atoms[index].np_pos,
                            frames,
                            i,
                        )

            # these copy calls take up a very long time
            system_array.append(system.copy(alt=False))
            mout.progress(i + frame_padding, frames + 2 * frame_padding)

        mout.progress(frames + 2 * frame_padding, frames + 2 * frame_padding)
        return system_array

    else:

        assert indices is not None
        assert len(indices) == 2
        assert frames <= 10
        assert not smooth
        assert ratios is None

        """

        AA AB AC
        BA BB BC
        CA CB CC

        """

        import numpy as np

        system_array = np.empty(
            (2 * frame_padding + frames, 2 * frame_padding + frames), dtype=object
        )
        # names_array = np.empty((2*frame_padding+frames,2*frame_padding+frames),dtype=str)

        iterange = range(-frame_padding, frames + frame_padding)

        system_array[frame_padding][frame_padding] = input1.copy(alt=False)
        system_array[-1 - frame_padding][-1 - frame_padding] = input2.copy(alt=False)

        corner1 = input1.copy(alt=False)
        corner2 = input1.copy(alt=False)

        corner1.atoms[indices[0]].position = input2.atoms[indices[0]].position
        corner2.atoms[indices[1]].position = input2.atoms[indices[1]].position

        # naming scheme
        from string import ascii_uppercase

        names = []
        for i in range(frame_padding):
            names.append(ascii_uppercase[-frame_padding + i])
        for i in range(frames):
            names.append(f"{i}")
        for i in range(frame_padding):
            names.append(ascii_uppercase[i])

        for i in iterange:
            for j in iterange:

                if system_array[i][j] is None:
                    system = input1.copy()

                    atom1 = system.atoms[indices[0]]
                    atom1.position = simple_interpolate(
                        input1.atoms[indices[0]].np_pos,
                        corner1.atoms[indices[0]].np_pos,
                        frames,
                        i,
                    )
                    atom2 = system.atoms[indices[1]]
                    atom2.position = simple_interpolate(
                        input1.atoms[indices[1]].np_pos,
                        corner2.atoms[indices[1]].np_pos,
                        frames,
                        j,
                    )

                    system_array[i][j] = system

                system_array[i][j].name = f"{names[i]}{names[j]}"

        return system_array


def custom_interpolate(start, end, ratio):
    return start + ratio * (end - start)


def simple_interpolate(start, end, frames, i):
    return start + i * (end - start) / (frames - 1)


def smooth_interpolate(start, end, frames, i):
    import math

    angle = simple_interpolate(0.0, math.pi, frames, i)
    return start + 0.5 * (1 - math.cos(angle)) * (end - start)


def auto_rotate(atoms):
    from .system import System

    if isinstance(atoms, System):
        sys = atoms
        atoms = auto_rotate(sys.ase_atoms)
        sys.set_coordinates(atoms)
        return sys

    atoms = atoms.copy()

    positions = atoms.get_positions()

    positions = [
        [p[0] for p in positions],
        [p[1] for p in positions],
        [p[2] for p in positions],
    ]

    import numpy as np
    from numpy.linalg import svd

    # fit a plane to the atomic positions
    points = np.reshape(
        positions, (np.shape(positions)[0], -1)
    )  # Collapse trialing dimensions
    assert (
        points.shape[0] <= points.shape[1]
    ), "There are only {} points in {} dimensions.".format(
        points.shape[1], points.shape[0]
    )
    central_point = points.mean(axis=1)
    x = points - central_point[:, np.newaxis]
    M = np.dot(x, x.T)  # Could also use np.cov(x) here.
    normal_vector = svd(M)[0][:, -1]

    # rotate the normal onto the Z-axis
    atoms.rotate(normal_vector, "z")

    # rotate the position vector of the first atom onto the x-axis
    positions = atoms.get_positions()
    index0_vector = np.array(
        [np.dot(positions[0], [1, 0, 0]), np.dot(positions[0], [0, 1, 0]), 0]
    )
    index1_vector = np.array(
        [np.dot(positions[1], [1, 0, 0]), np.dot(positions[1], [0, 1, 0]), 0]
    )
    # index0_vector = positions[0]
    # index1_vector = positions[1]

    if index0_vector[0] < 0:
        index0_vector = [index0_vector[0], index0_vector[1], 0]
    # print(index0_vector)
    atoms.rotate(index1_vector - index0_vector, "x")
    # print(atoms.get_positions()[0])

    return atoms



# import math


def find_barrier_stationary_points(
    xdata, ydata, yerr=None, reduction_order=2, show=False, tshicut=None, tslocut=None
):
    import numpy as np
    import scipy.signal as sps

    # find the x-location of the stationary points

    if yerr is None:
        yerr = [None for x in xdata]

    maxima_indices = sps.argrelextrema(np.array(ydata), np.greater)[0]
    minima_indices = sps.argrelextrema(np.array(ydata), np.less)[0]

    max_coords = []
    for i in maxima_indices:
        max_coords.append([xdata[i], ydata[i], yerr[i]])

    min_coords = []
    for i in minima_indices:
        min_coords.append([xdata[i], ydata[i], yerr[i]])

    if reduction_order == 2:

        maxima_ys = []
        for p in max_coords:
            valid = True
            if tshicut is not None and p[0] >= tshicut:
                valid = False
            elif tslocut is not None and p[0] <= tslocut:
                valid = False

            if valid:
                maxima_ys.append(p[1])
            else:
                maxima_ys.append(-1)

        # if tshicut is not None:
        # 	if tslocut is not None:
        # 		maxima_ys = [p[1] for p in max_coords if p[0] <= tshicut and p[0] >= tslocut]
        # 	else:
        # 		maxima_ys = [p[1] for p in max_coords if p[0] <= tshicut]
        # 		# maxima_ys = [p[1] for p in max_coords if p[0] >= tslocut]
        # else:
        # 	maxima_ys = [p[1] for p in max_coords]

        if not maxima_ys:
            import mout

            mout.warningOut("No maximas found!")
            # import matplotlib.pyplot as plt
            # print(min_coords)
            # plt.plot(xdata,ydata)
            # plt.scatter([c[0] for c in min_coords],[c[1] for c in min_coords])
            # plt.show()
            assert len(min_coords) == 1
            return min_coords, []

        highest_maxima_index = maxima_ys.index(max(maxima_ys))

        max_coords = [max_coords[highest_maxima_index]]

        maxima_x = max_coords[0][0]

        min_before = [p for p in min_coords if p[0] < maxima_x]
        min_after = [p for p in min_coords if p[0] > maxima_x]

        min_before = [p for p in min_before if p[1] == min([p[1] for p in min_before])]
        min_after = [p for p in min_after if p[1] == min([p[1] for p in min_after])]

        min_coords = min_before + min_after

    if show:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        plt.plot(xdata, ydata)

        plt.errorbar(
            [p[0] for p in max_coords],
            [p[1] for p in max_coords],
            yerr=[p[2] for p in max_coords],
            fmt="o",
        )
        plt.errorbar(
            [p[0] for p in min_coords],
            [p[1] for p in min_coords],
            yerr=[p[2] for p in min_coords],
            fmt="o",
        )

        plt.show()

    return min_coords, max_coords


# def eckartFit(xdata,ydata):
# 	return None


def sqrt(x):
    import numpy as np

    return np.power(x, 0.5)


def eckart(x, barrier, asymmetry, length, xts=0.0):
    import numpy as np

    if isinstance(x, list):
        x = np.array(x)

    reverse_barrier = barrier - asymmetry

    # print(reverse_barrier)

    A = barrier - reverse_barrier

    B = np.power(sqrt(reverse_barrier) + sqrt(barrier), 2)

    y = -np.exp(2.0 * np.pi * (x - xts) / length)

    V = -A * y * np.power(1.0 - y, -1) - B * y * np.power(1.0 - y, -2)

    return V



class Restraint:
    def __init__(self, atoms, type=None, weights=None, force_constant=None):

        self.atoms = atoms
        self.range = None
        self.force_constant = force_constant
        self.weights = weights
        self._values = None

        if type == None:
            self.type = self.determine_type()

        self.name = self.determine_name()

    def determine_type(self):
        import mout

        """ Restraint types:
                Two-Atom: Distance							Implemented
                Three-Atom: Angle							Implemented
                Four-Atom: Torsional / J-Coupling
                Four-Atom: Generalised Distance (Transfer)	Implemented
                Five-Atom: Plane-Point
                Six-Atom: Generalised Distance
                Eight-Atom: Plane-Plane						Implemented
        """

        if len(self.atoms) == 2:
            return "distance"

        elif len(self.atoms) == 3:
            return "angle"

        elif len(self.atoms) == 4:
            assert len(self.weights) == 2
            return "transfer"

        elif len(self.atoms) == 8:
            return "plane-plane"

        else:
            mout.errorOut("Could not determine restraint type", fatal=True)

    def determine_name(self):
        if self.type == "distance":
            return "dist(" + self.atoms[0].name + "-" + self.atoms[1].name + ")"
        elif self.type == "transfer":
            return (
                "RC("
                + self.atoms[0].name
                + "-"
                + self.atoms[1].name
                + ","
                + self.atoms[2].name
                + "-"
                + self.atoms[3].name
                + ")"
            )
        elif self.type == "angle":
            return (
                "ang("
                + self.atoms[0].name
                + ","
                + self.atoms[1].name
                + ","
                + self.atoms[2].name
                + ")"
            )
        elif self.type == "plane-plane":
            return "plane-plane(...)"
        else:
            return "restraint"

    def value(self, system, scale_angle=None):
        import numpy as np
        import mout

        this_atoms = []
        for atom in self.atoms:
            this_atom = system.atoms[atom.index]
            assert str(this_atom) == str(atom)
            this_atoms.append(this_atom)

        if self.type == "distance":
            return np.linalg.norm(this_atoms[0].np_pos - this_atoms[1].np_pos)

        elif self.type == "transfer":
            dist1 = np.linalg.norm(this_atoms[0].np_pos - this_atoms[1].np_pos)
            dist2 = np.linalg.norm(this_atoms[2].np_pos - this_atoms[3].np_pos)
            return self.weights[0] * dist1 + self.weights[1] * dist2

        elif self.type == "angle":
            vec1 = this_atoms[0].np_pos - this_atoms[1].np_pos
            vec2 = this_atoms[2].np_pos - this_atoms[1].np_pos
            dot = np.dot(vec1, vec2)
            norms = np.linalg.norm(vec1) * np.linalg.norm(vec2)
            if scale_angle == None:
                return np.arccos(dot / norms) / np.pi * 180
            else:
                return np.arccos(dot / norms) / np.pi * scale_angle

        elif self.type == "plane-plane":
            # (r1 - r2) × (r3 - r4) and (r5 - r6) × (r7 - r8)
            vec1 = this_atoms[0].np_pos - this_atoms[1].np_pos
            vec2 = this_atoms[2].np_pos - this_atoms[3].np_pos
            vec3 = this_atoms[4].np_pos - this_atoms[5].np_pos
            vec4 = this_atoms[6].np_pos - this_atoms[7].np_pos
            cross1 = np.cross(vec1, vec2)
            cross2 = np.cross(vec3, vec4)
            dot = np.dot(cross1, cross2)
            norms = np.linalg.norm(cross1) * np.linalg.norm(cross2)
            if scale_angle == None:
                return np.arccos(dot / norms) / np.pi * 180
            else:
                return np.arccos(dot / norms) / np.pi * scale_angle

        else:
            mout.errorOut("Unsupported restraint type", fatal=True)

    def values(self, system_list=None, scale_angle=None):
        if system_list is None:
            return self._values
        else:
            from .system import System

            assert isinstance(system_list, list)
            values = []
            for system in system_list:
                assert isinstance(system, System)
                values.append(self.value(system, scale_angle=scale_angle))
            self._values = values
            return values

    def equi_values(self, n):
        assert self._values is not None

        from scipy.interpolate import interp1d

        x = [i for i in range(len(self._values))]

        inter_func = interp1d(x, self._values)

        equi = []

        for i in range(n):
            this_x = x[0] + i * (x[-1] - x[0]) / (n - 1)

            equi.append(inter_func(this_x))

        return equi

    def set_values(self, values):
        self._values = values

    def fetch_atoms(self, new_system, new_residues, res_map=None, verbosity=1):

        import mout

        count = 0

        new_atoms = []

        for atom in self.atoms:

            found = False

            # print(atom.residue,new_residues)

            if res_map is not None:
                for pair in res_map:
                    # if atom.residue in pair[0]:
                    # 	print(atom.residue,pair[0])
                    # 	atom.residue = pair[1]
                    # 	print(atom.residue,pair[0])
                    # 	found_map = True
                    # 	break
                    if atom.residue == pair[1]:
                        # print(atom.residue)
                        atom.residue = pair[0]
                        # print(atom.residue)
                        break

            for res in new_residues:
                if res.name == atom.residue:
                    new_atoms.append(res.get_atom(atom.name))
                    found = True
                    count += 1

            if not found:
                mout.varOut("atoms", self.atoms)
                mout.varOut("new_residues", new_residues)
                mout.errorOut(
                    "Could not find match for "
                    + atom.name
                    + " ("
                    + atom.residue
                    + ")"
                    + " in residues "
                    + str(new_residues),
                    fatal=True,
                    code="Restraint.fetch_atoms",
                )

        if verbosity > 0:
            mout.out(
                "Retrieved "
                + str(count)
                + " atoms from "
                + new_system.name
                + "'s "
                + str(new_residues)
            )

        self.atoms = new_atoms

    def amber_block(self, index, harmonic_width=10.0, debug=False):

        value = self.values()[index]

        end = "\n"

        rst_buffer = "&rst" + end

        rst_buffer += "iat="

        for atom in self.atoms:
            rst_buffer += str(atom.pdb_index) + ","

        if debug:
            rst_buffer += " #"

            for atom in self.atoms:
                rst_buffer += str(atom.name) + ", "

        rst_buffer += end

        if self.type == "transfer":
            rst_buffer += "rstwt="
            rst_buffer += str(self.weights[0]) + ","
            rst_buffer += str(self.weights[1]) + "," + end

        r1 = value - harmonic_width
        r2 = value
        r3 = value
        r4 = value + harmonic_width

        rk2 = self.force_constant
        rk3 = self.force_constant

        rst_buffer += "r1=" + str(r1) + "," + end
        rst_buffer += "r2=" + str(r2) + "," + end
        rst_buffer += "r3=" + str(r3) + "," + end
        rst_buffer += "r4=" + str(r4) + "," + end
        rst_buffer += "rk2=" + str(rk2) + "," + end
        rst_buffer += "rk3=" + str(rk3) + "," + end

        rst_buffer += "/" + end

        return rst_buffer

    def copy(self):
        import copy

        return copy.deepcopy(self)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name



from .features import (
    features_from_group,
    Feature,
    features_from_mol,
    raw_features_from_mol,
)

from .draw import *
from .mol import *
from .fingerprint import get_similar
from .tools import sdf_align, sdf_combine
from .sucos import SuCOS_score
from .classify import classify_mol
from .compare import compare_mols



from collections import UserList


class NamedList(UserList):
    """Subclass of collections.UserList which allows for subscription using object names and numbers according to

    key: int 		returns list[int]
    key: 'n'+int 	returns list[obj.number == int]
    key: 't'+str 	returns list[obj.type == str]
    key: str 		returns list[obj.name == str]
    key: str' 'int  returns list[obj.name == str and obj.index == int]
    key: str' n'int  returns list[obj.name == str and obj.number == int]

    Recursion is allowed via: list[[key1,key2,key3]] --> [list[key1],list[key2],list[key3]]

    """

    def __init__(self, inherited_list=[]):
        super(NamedList, self).__init__(inherited_list)

    def get_matches(self, name=None, number=None):

        assert any([name, number is not None])

        # name only
        if name and number is None:
            matches = [obj for obj in self.data if obj.name == name]
            return matches

        # name and number
        elif name and number is not None:
            matches = [
                obj for obj in self.data if obj.number == number and obj.name == name
            ]

            if not matches:
                return None
            elif len(matches) > 1:
                import mout

                mout.error(f"{self}.get_matches({name=}, {number=})")
                mout.error("Multiple matches to a supposedly unique query!", fatal=True)
            else:
                return matches[0]

        # number only
        elif not name and number is not None:
            matches = [obj for obj in self.data if obj.number == number]

            if not matches:
                return None
            elif len(matches) > 1:
                import mout

                mout.error(f"{self}.get_matches({name=}, {number=})")
                mout.error("Multiple matches to a supposedly unique query!", fatal=True)
            else:
                return matches[0]

    def __getitem__(self, key):

        # if key is a slice
        if isinstance(key, slice):
            return NamedList(self.data[key])

        # if key is a list
        if isinstance(key, list) or isinstance(key, tuple) or isinstance(key, set):

            data = []
            for item in key:
                data.append(self.__getitem__(item))

            return NamedList(data)

        # try the key as an integer index
        try:
            key = int(key)
            return self.data[key]

        except ValueError:

            # use
            if key.startswith("n"):

                try:
                    number = int(key[1:])
                except ValueError:
                    import mout

                    mout.error(f"Could not convert {key[1:]} to int")
                    mout.error(
                        "Lookup key should be in the form 'n'INDEX for a specific lookup"
                    )
                    raise ValueError("Could not convert {key[1:]} to int")

                # find matches
                matches = self.get_matches(number=number)
                return matches

            # get by type
            if key.startswith("t"):
                type_str = key[1:]
                return [d for d in self.data if d.type == type_str]

            # key may contain space-separated values
            split_key = key.split(" ")

            # if just a word return all matches
            if len(split_key) == 1:
                return self.get_matches(name=split_key[0])

            # if two words
            elif len(split_key) == 2:

                # first word should be a name
                name = split_key[0]

                if split_key[1].startswith("n"):
                    try:
                        number = int(split_key[1][1:])
                    except ValueError:
                        import mout

                        mout.error(f"Could not convert {key[1:]} to int")
                        mout.error(
                            "Lookup key should be in the form NAME' n'INDEX for a specific lookup"
                        )
                        return None

                    # find matches
                    matches = self.get_matches(name=name, number=number)
                    return matches

                else:

                    # second word should be an index
                    try:
                        index = int(split_key[1])
                    except ValueError:
                        import mout

                        mout.error(f"Could not convert {split_key[1]} to int")
                        mout.error(
                            "Lookup key should be in the form NAME' 'INDEX for a specific lookup"
                        )
                        return None

                match = self.data[index]

                assert match.name == name

                return match

            else:
                import mout

                mout.error("key must be an integer, object name or 'obj_name index'")
                return None



from .mol import mol_from_smiles
from rdkit import DataStructs
from rdkit.Chem.Fingerprints import FingerprintMols


def get_similar(mols, query, index=False, score=False, threshold=1.0):
    if isinstance(query, str):
        query = mol_from_smiles(query)

    kwargs = dict(
        minPath=1,
        maxPath=7,
        fpSize=2048,
        bitsPerHash=2,
        useHs=True,
        tgtDensity=0.0,
        minSize=128,
    )

    fp = FingerprintMols.FingerprintMol(query, **kwargs)
    fps = [FingerprintMols.FingerprintMol(m, **kwargs) for m in mols]

    scores = DataStructs.BulkTanimotoSimilarity(fp, fps)

    if index:
        return [i for i, s in enumerate(scores) if s >= threshold]

    elif not index and score:
        return [(m, s) for m, s in zip(mols, scores) if s >= threshold]

    else:
        return [m for m, s in zip(mols, scores) if s >= threshold]



import plotly.graph_objects as go
import mgo
import numpy as np
from .rand import random_point_spherical


class CompoundVolume:
    def __init__(self, reference=None):
        self.volumes = []
        self.reference = reference

    # def cull_overlapping_volumes(self):

    def add_volume(self, volume):
        volume.index = self.num_volumes
        self.volumes.append(volume)

    def plot(self, distance, fig=None, show=False, samples=20):
        if fig is None:
            fig = go.Figure()

        for vol in self.volumes:
            fig = vol.plot(distance=distance, samples=samples, fig=fig, show=False)

        if show:
            fig.show()

        return fig

    def is_inside(self, point):

        for vol in self.volumes:

            if vol.is_inside(point):
                return True
        return False

    def is_within(self, point, within):
        for vol in self.volumes:
            if vol.is_within(point, within):
                return True
        return False

    @property
    def num_volumes(self):
        return len(self.volumes)

    def simplify(self):

        del_list = []

        for this_volume in self.volumes:
            for other_volume in self.volumes:

                if this_volume.index == other_volume.index:
                    continue

                if this_volume in del_list:
                    continue

                if other_volume in del_list:
                    continue

                print(this_volume, this_volume.index, other_volume, other_volume.index)

                if this_volume.always_inside(other_volume):
                    del_list.append(this_volume)

        print(del_list)

        for volume in reversed(self.volumes):
            if volume in del_list:
                del volume


class Sphere:
    def __init__(self, centre, radius, index=None):
        self.index = index
        self.centre = np.array(centre)
        self.radius = np.array(radius)

    def is_inside(self, point):
        return np.linalg.norm(point - self.centre) < self.radius


import mout


class CappedCone:
    def __init__(self, origin, target, radius, name=None, index=None):

        # parameters
        self.name = name
        self.origin = np.array(origin)
        self.target = np.array(target)
        self.radius = radius
        self.index = index

        # derived internals
        self.vec_origin_target = self.target - self.origin
        self.dist_origin_target = np.linalg.norm(self.vec_origin_target)
        self.unit_origin_target = self.vec_origin_target / self.dist_origin_target
        self.theta = np.arctan(self.radius / self.dist_origin_target)
        self.cos_theta = np.cos(self.theta)
        self.start = self.dist_origin_target * self.cos_theta * self.cos_theta

        # child volumes
        self.cap_radius = self.dist_origin_target * np.sin(self.theta)
        self.cap_sphere = Sphere(target, self.cap_radius)

        # plotting variables
        self.vec_origin_point = None

        self._expanded_volume = None

    def is_inside(self, point):

        self._is_inside_cap = self.cap_sphere.is_inside(point)

        if self._is_inside_cap:
            return True

        self.vec_origin_point = point - self.origin
        self.dist_origin_point = np.linalg.norm(self.vec_origin_point)

        if self.dist_origin_point == 0:
            return False

        self.unit_origin_point = self.vec_origin_point / self.dist_origin_point

        if self.dist_origin_point < self.start:
            return False

        self.origin_point_dot_origin_target = np.dot(
            self.unit_origin_point, self.unit_origin_target
        )

        if self.origin_point_dot_origin_target > self.cos_theta:
            # print(f'{self.origin_point_dot_origin_target=}')
            # print(f'{self.cos_theta=}')
            return True

        return False

    def is_within(self, point, within):

        if self._expanded_volume is None or self._expanded_volume._expansion != within:
            new_origin = self.origin - self.unit_origin_target * within / np.sin(
                self.theta
            )

            self._expanded_volume = CappedCone(
                origin=new_origin,
                target=self.target,
                radius=self.radius + within,
                name=self.name + " expanded",
            )
            self._expanded_volume._expansion = within

        return self._expanded_volume.is_inside(point)

    def test_random_point(self, n=1, distance=5):

        fig = self.plot(distance)

        for i in range(n):

            # point = random_point(self.origin,distance)
            point = np.array([-2.127, 2.62, -17.7])

            # print(f'{point=}')

            is_inside = self.is_inside(point)

            # print(f'{is_inside=}')
            # print(f'{self._is_inside_cap=}')

            if self.vec_origin_point is not None:
                trace = mgo.vector_trace(
                    self.origin, self.vec_origin_point, name=f"{is_inside=}"
                )
                fig.add_trace(trace)

        fig.show()

    def plot(self, distance, samples=20, fig=None, show=False, verbosity=1):

        if fig is None:
            fig = go.Figure()

        trace = mgo.cone_trace(
            self.origin,
            self.target,
            self.radius,
            distance,
            samples=samples,
            name=self.name,
            start_at_target=True,
            plot_caps=True,
            verbosity=verbosity - 1,
        )
        fig.add_trace(trace)

        if show:
            fig.show()

        return fig

    @mout.debug_log
    def always_inside(self, other):

        if isinstance(other, CappedCone):

            assert np.linalg.norm(self.origin - other.origin) == 0

            if (
                other.dist_origin_target - other.cap_radius
                < self.dist_origin_target - self.cap_radius
            ):
                print("case1")
                return False

            # if not other.is_inside(self.target):
            # 	print(f'{self}.target is not inside {other}')
            # 	return False

            # print(f'{self}.target is inside {other}')

            if other.theta < self.theta:
                print("case2")
                return False

            phi = np.arccos(np.dot(self.unit_origin_target, other.unit_origin_target))

            if other.theta < phi + self.theta:
                print("case3")
                return False

            return True

        else:
            raise Exception(
                f"Unsupported: CappedCone.always_inside(self,type({other}))"
            )

    def __repr__(self):
        if self.name:
            return self.name
        else:
            return f"CappedCone index={self.index}"



from ase.data import chemical_symbols

chemical_symbols = chemical_symbols[1:]
chemical_symbols = sorted(chemical_symbols, key=lambda x: len(x), reverse=True)


def formula_to_atomtype_dict(formula):
    """

    :param formula:

    """

    import re

    remaining = formula.replace("+", "").replace("-", "")

    atomtype_dict = {}

    while remaining:
        for symbol in chemical_symbols:
            if remaining.startswith(symbol):
                remaining = remaining.removeprefix(symbol)

                count_str = re.findall(r"^([0-9]*)", remaining)[0]

                if count_str:
                    remaining = remaining.removeprefix(count_str)
                    count = int(count_str)
                else:
                    count = 1

                atomtype_dict[symbol] = count

                break
        else:
            raise Exception(
                f"Could not match any chemical symbol to the start of {remaining} {formula=}"
            )

    return atomtype_dict


def combine_atomtype_dicts(atomtype_dicts):
    """

    :param atomtype_dicts:

    """

    merged = {}
    for atomtype_dict in atomtype_dicts:
        for symbol, count in atomtype_dict.items():
            if symbol not in merged:
                merged[symbol] = 0

            merged[symbol] += count

    return merged


def atomtype_dict_to_formula(atomtype_dict):
    """

    :param atomtype_dict:

    """

    symbols = []

    """For organic compounds, the order is carbon, hydrogen,
	then all other elements in alphabetical order of their chemical symbols."""

    key = "C"
    if key in atomtype_dict:
        value = atomtype_dict[key]
        if value > 1:
            symbols.append(f"C{value}")
        else:
            symbols.append("C")

    key = "H"
    if key in atomtype_dict:
        value = atomtype_dict[key]
        if value > 1:
            symbols.append(f"{key}{value}")
        else:
            symbols.append(key)

    keys = sorted(list(atomtype_dict.keys()))

    for key in keys:
        value = atomtype_dict[key]
        if key in ["C", "H"]:
            continue
        if value > 1:
            symbols.append(f"{key}{value}")
        else:
            symbols.append(key)

    return "".join(symbols)


def subtract_atomtype_dict(d1, d2, ignore_hydrogen=False):

    all_keys = set(d1.keys()) | set(d2.keys())

    d = {}

    for key in all_keys:

        if ignore_hydrogen and key == "H":
            continue

        n1 = d1[key] if key in d1 else 0
        n2 = d2[key] if key in d2 else 0

        n = n1 - n2

        if n > 0:
            d[key] = n

    return d



#!/usr/bin/env python3

import mout
import molparse as mp
from mout.testing import TestStatus
from molparse.amino import INTERACTION_SITES, BB_NAMES, RES_NAMES, BB_INTERACTION_SITES

import os
import glob
from pprint import pprint


@mout.debug_log
def check_bb_interactions():
    test_status = TestStatus()

    ### Backbone interactions

    for atom_name in BB_NAMES:

        # get all interactions for those names
        interactions = get_interactions_by_atom_name(atom_name)
        if interactions:
            mout.var(f"interactions({atom_name})", interactions.keys())

        # warn if BB interactions present
        if interactions:
            test_status.warning(
                'BB interactions present, move to AMINO_INTERACTIONS["BB"]'
            )

    return test_status


@mout.debug_log
def check_interaction_source():
    test_status = TestStatus()

    for res_name, interactions in INTERACTION_SITES.items():

        for interaction in interactions:

            if "source" not in interaction:
                test_status.error(
                    f'No source field in {res_name}: {interaction["type"]} {interaction["atoms"]}'
                )
                continue

            if "unsure" in interaction["source"]:
                test_status.warning(
                    f'unsure source for {res_name}: {interaction["type"]} {interaction["atoms"]}'
                )

    return test_status


@mout.debug_log
def check_interaction_pair(int1, int2):
    test_status = TestStatus()

    # for each interaction in an interaction pair check that there is another interaction with the same atoms

    for res_name, interactions in INTERACTION_SITES.items():

        for interaction in interactions:

            if interaction["type"] == int1:
                matches = [
                    i
                    for i in interactions
                    if i["type"] == int2 and i["atoms"] == interaction["atoms"]
                ]
                if len(matches) != 1:
                    test_status.warning(
                        f'{res_name}: No {int2} for {interaction["atoms"]}'
                    )

            elif interaction["type"] == int2:
                matches = [
                    i
                    for i in interactions
                    if i["type"] == int1 and i["atoms"] == interaction["atoms"]
                ]
                if len(matches) != 1:
                    test_status.warning(
                        f'{res_name}: No {int1} for {interaction["atoms"]}'
                    )

    return test_status


@mout.debug_log
def check_pi_stack_pair():
    test_status = TestStatus()

    # for each interaction in an interaction pair check that there is another interaction with the same atoms

    for res_name, interactions in INTERACTION_SITES.items():

        for interaction in interactions:

            # every pi_stack should be able to do a pi_cation
            if interaction["type"] == "pi_stacking":
                matches = [
                    i
                    for i in interactions
                    if i["type"] == "pi_cation" and i["atoms"] == interaction["atoms"]
                ]
                if len(matches) != 1:
                    test_status.warning(
                        f'{res_name}: No pi_cation for {interaction["atoms"]}'
                    )

            # only pi_cation interactions with the ring on the protein infer pi_stacking
            elif interaction["type"] == "pi_cation":
                if len(interaction["atoms"]) < 4:
                    continue
                matches = [
                    i
                    for i in interactions
                    if i["type"] == "pi_stacking" and i["atoms"] == interaction["atoms"]
                ]
                if len(matches) != 1:
                    test_status.warning(
                        f'{res_name}: No pi_stacking for {interaction["atoms"]}'
                    )

    return test_status


def get_interactions_by_atom_name(atom_name):
    result = {}

    for res_name, interactions in INTERACTION_SITES.items():

        for interaction in interactions:

            if atom_name in interaction["atoms"]:

                int_type = interaction["type"]

                if int_type not in result:
                    result[int_type] = []

                if res_name not in result[int_type]:
                    result[int_type].append(res_name)

    return result


@mout.debug_log
def check_amino_interaction_sites_method():
    test_status = TestStatus()

    ref_dir = f"{os.path.dirname(__file__)}/../molparse/ref"

    for res_name in RES_NAMES:

        files = glob.glob(f"{ref_dir}/{res_name}.pdb")

        if len(files) == 0:
            test_status.error(f"no reference file for {res_name}")
            continue

        elif len(files) == 0:
            test_status.error(f"multiple reference files for {res_name}")
            continue

        system = mp.parse(files[0], verbosity=0)

        interactions = system["r0"].interaction_sites

        if not interactions:
            test_status.warning(f"No interactions for {res_name}")

    return test_status


def main():
    mout.header(__file__)

    # used to generate INTERACTION_SITES['BB']
    check_bb_interactions()

    check_interaction_source()
    check_interaction_pair("hydrogen_donor", "water_donor")
    check_interaction_pair("hydrogen_acceptor", "water_acceptor")
    check_pi_stack_pair()

    # check that histidine interactions are identical?

    check_amino_interaction_sites_method()

    TestStatus.summary()


if __name__ == "__main__":
    main()



from collections import UserDict


def parseNDX(file, verbosity=1):
    import mout
    import mcol

    if verbosity:
        mout.out(f"parsing {mcol.file}{file}{mcol.clear} ... ", end="")

    index_dict = GromacsIndex()

    with open(file) as f:

        group_name = None
        group_indices = None

        for line in f:

            line = line.strip()

            if line.startswith("["):

                if group_name:
                    index_dict[group_name] = group_indices

                group_name = line.replace("[", "").replace("]", "").strip()
                group_indices = []

            elif group_name:

                for word in line.split(" "):

                    if not word:
                        continue

                    group_indices.append(int(word.strip()))

            else:
                print(line)

            if group_name:
                index_dict[group_name] = group_indices

    if verbosity:
        mout.out(f"Done.")

    return index_dict


def writeNDX(file, ndx, chunk_size=15, verbosity=1):
    import mout
    import mcol

    if verbosity:
        mout.out(f"writing {mcol.file}{file}{mcol.clear} ... ", end="")

    with open(file, "wt") as f:

        for group_name in ndx:

            f.write(f"[ {group_name} ]\n")

            indices = ndx[group_name]

            if len(indices) < chunk_size:
                indices = [str(i) for i in indices]
                f.write(f'{" ".join(indices)}\n')

            else:
                for index_chunk in grouper(indices, chunk_size, incomplete="ignore"):
                    index_chunk = [str(i) for i in index_chunk]
                    f.write(f'{" ".join(index_chunk)}\n')

    if verbosity:
        mout.out(f"Done.")


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def grouper(iterable, n, *, incomplete="fill", fillvalue=None):
    """Collect data into non-overlapping fixed-length chunks or blocks

    * grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    * grouper('ABCDEFG', 3, incomplete='strict') --> ABC DEF ValueError
    * grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF

    """

    args = [iter(iterable)] * n
    if incomplete == "fill":
        return zip_longest(*args, fillvalue=fillvalue)
    if incomplete == "strict":
        return zip(*args, strict=True)
    if incomplete == "ignore":
        return zip(*args)
    else:
        raise ValueError("Expected fill, strict, or ignore")


class GromacsIndex(UserDict):
    def __init__(self):
        super(GromacsIndex, self).__init__()

    @property
    def group_names(self):
        return self.data.keys()

    def shift(self, shift=-1):
        for key in self.data:
            self.data[key] = [i + shift for i in self.data[key]]

    def summary(self):

        import mout
        import mcol

        for group in self.data:
            print(
                f"{mcol.varType}Group {mcol.arg}{group}{mcol.clear} contains {mcol.result}{len(self.data[group])}{mcol.clear} indices"
            )

    def write(self, file):
        writeNDX(file, self)



# implementation of SuCOS from https://github.com/susanhleung/SuCOS https://doi.org/10.26434/chemrxiv.8100203.v1

import os
from numpy import clip

from rdkit import RDConfig
from rdkit.Chem import AllChem, rdShapeHelpers
from rdkit.Chem.FeatMaps import FeatMaps
from rdkit.Chem.rdmolops import CombineMols
import mrich
from pandas import DataFrame

# feature setup

feature_factory = AllChem.BuildFeatureFactory(
    os.path.join(RDConfig.RDDataDir, "BaseFeatures.fdef")
)

feature_map_params = {}
for k in feature_factory.GetFeatureFamilies():
    feature_params = FeatMaps.FeatMapParams()
    feature_map_params[k] = feature_params

keep = (
    "Donor",
    "Acceptor",
    "NegIonizable",
    "PosIonizable",
    "ZnBinder",
    "Aromatic",
    "Hydrophobe",
    "LumpedHydrophobe",
)


def feature_map_score(
    inspiration,
    derivative,
    score_mode=FeatMaps.FeatMapScoreMode.All,
    debug=False,
    draw=False,
    return_data: bool = False,
):

    inspiration_features = [
        f
        for f in feature_factory.GetFeaturesForMol(inspiration)
        if f.GetFamily() in keep
    ]
    derivative_features = [
        f
        for f in feature_factory.GetFeaturesForMol(derivative)
        if f.GetFamily() in keep
    ]

    if draw:
        from .draw import draw_mol

        mrich.h3("inspiration")
        draw_mol(inspiration, feats=inspiration_features)
        mrich.h3("derivative")
        draw_mol(derivative, feats=derivative_features)

    inspiration_feature_map = FeatMaps.FeatMap(
        feats=inspiration_features,
        weights=[1] * len(inspiration_features),
        params=feature_map_params,
    )
    derivative_feature_map = FeatMaps.FeatMap(
        feats=derivative_features,
        weights=[1] * len(derivative_features),
        params=feature_map_params,
    )

    inspiration_feature_map.scoreMode = score_mode

    feature_score = inspiration_feature_map.ScoreFeats(derivative_features)

    score_vector = [0] * inspiration_feature_map.GetNumFeatures()

    inspiration_feature_map.ScoreFeats(derivative_features, mapScoreVect=score_vector)

    if return_data or debug:
        inspiration_data = []
        for feature, score in zip(inspiration_features, score_vector):
            inspiration_data.append(
                dict(
                    family=feature.GetFamily(),
                    type=feature.GetType(),
                    atom_ids=str(feature.GetAtomIds()),
                    score=score,
                )
            )

    if debug:

        from rich.table import Table

        mrich.h3("Debug")

        if draw:
            from .draw import draw_flat

            display(draw_flat(inspiration, indices=True))
            display(draw_flat(derivative, indices=True))

        table = Table(title="inspiration features")
        table.add_column("family")
        table.add_column("type")
        table.add_column("atom_ids")
        table.add_column("recapitulation score")

        for d in inspiration_data:
            table.add_row(
                d["family"], d["type"], str(d["atom_ids"]), f"{d['score']:.3f}"
            )
        mrich.print(table)

        table = Table(title="derivative features")
        table.add_column("family")
        table.add_column("type")
        table.add_column("atom_ids")
        # table.add_column("recapitulation score")

        for feature in derivative_features:
            # mrich.debug(feature.GetFamily(), feature.GetType(), feature.GetAtomIds(), score)
            table.add_row(
                feature.GetFamily(), feature.GetType(), str(feature.GetAtomIds())
            )
        mrich.print(table)

        # mrich.print(dir(feature.GetPos()))

    feature_score /= min(
        inspiration_feature_map.GetNumFeatures(), len(derivative_features)
    )

    if return_data:
        return feature_score, inspiration_data

    return feature_score


def multi_feature_map_score(
    inspirations,
    derivative,
    draw=False,
    debug=False,
    score_mode=FeatMaps.FeatMapScoreMode.All,
):

    inspiration_feature_lists = [
        [f for f in feature_factory.GetFeaturesForMol(x) if f.GetFamily() in keep]
        for x in inspirations
    ]
    derivative_features = [
        f
        for f in feature_factory.GetFeaturesForMol(derivative)
        if f.GetFamily() in keep
    ]

    combined_inspiration_features = sum(inspiration_feature_lists, [])

    if draw:
        from .draw import draw_mol

        for inspiration, features in zip(inspirations, inspiration_feature_lists):
            draw_mol(inspiration, feats=features)

        draw_mol(derivative, feats=combined_inspiration_features)

        draw_mol(derivative, feats=derivative_features)

    derivative_feature_map = FeatMaps.FeatMap(
        feats=derivative_features,
        weights=[1] * len(derivative_features),
        params=feature_map_params,
    )

    # loop over inspirations

    score_dict = {}

    for i, inspiration in enumerate(inspirations):

        score_dict[i] = {}

        if debug:
            mrich.title(inspiration)

        inspiration_features = [
            f
            for f in feature_factory.GetFeaturesForMol(inspiration)
            if f.GetFamily() in keep
        ]
        inspiration_feature_map = FeatMaps.FeatMap(
            feats=inspiration_features,
            weights=[1] * len(inspiration_features),
            params=feature_map_params,
        )

        score_vector = [0] * derivative_feature_map.GetNumFeatures()

        feature_score = derivative_feature_map.ScoreFeats(
            inspiration_features, mapScoreVect=score_vector
        )

        for j, (feature, score) in enumerate(zip(derivative_features, score_vector)):
            if debug:
                mrich.var(feature.GetFamily(), score)

            score_dict[i][j] = score

    i_list = list(score_dict.keys())

    combined_score_vector = []

    for j, feature in enumerate(derivative_features):

        scores = [score_dict[i][j] for i in i_list]

        best_score = max(scores)

        combined_score_vector.append(best_score)

        # print(feature.GetFamily(), scores, best_score)

    feature_score = sum(combined_score_vector) / len(derivative_features)

    return feature_score


def SuCOS_score(
    inspiration,
    derivative,
    recapitulation_threshold: float = 0.6,
    return_all=False,
    print_scores=False,
    debug: bool = False,
    draw: bool = False,
    **kwargs,
):

    if isinstance(inspiration, list):
        if len(inspiration) == 1:
            multi = False
            inspiration = inspiration[0]
        else:
            multi = True
    else:
        multi = False

    if multi:
        feature_score = multi_feature_map_score(
            inspiration, derivative, debug=debug, draw=draw, **kwargs
        )
        inspiration_data = []
    else:
        feature_score, inspiration_data = feature_map_score(
            inspiration, derivative, debug=debug, draw=draw, return_data=True, **kwargs
        )

    if debug or print_scores or return_all:
        df = DataFrame(inspiration_data)
        df = df.groupby("atom_ids").max("score")
        recapitulation_count = len(df[df["score"] > recapitulation_threshold])
        recapitulation_fraction = recapitulation_count / len(df)

    feature_score = clip(feature_score, 0, 1)

    if multi:

        mol = inspiration.pop()

        while inspiration:
            mol = CombineMols(mol, inspiration.pop())

        protrude_dist = rdShapeHelpers.ShapeProtrudeDist(
            derivative, mol, allowReordering=False
        )

    else:
        protrude_dist = rdShapeHelpers.ShapeProtrudeDist(
            inspiration, derivative, allowReordering=False
        )

    protrude_dist = clip(protrude_dist, 0, 1)
    volume_score = 1 - protrude_dist

    SuCOS_score = (feature_score + volume_score) * 0.5

    if debug or print_scores:
        if debug:
            mrich.h3("scores")
        mrich.var("feature_score", feature_score)
        mrich.var("volume_score", volume_score)
        mrich.var("average_score", SuCOS_score)
        mrich.var("recapitulation_count", recapitulation_count)

    if return_all:
        return dict(
            average_score=SuCOS_score,
            feature_score=feature_score,
            volume_score=volume_score,
            recapitulation_count=recapitulation_count,
            recapitulation_fraction=recapitulation_fraction,
        )
    else:
        return SuCOS_score



import py3Dmol
from rdkit import Chem
from rdkit.Chem.Draw import IPythonConsole, rdMolDraw2D
from rdkit.Chem import rdFMCS, AllChem, Draw
from rdkit.Chem.Features.ShowFeats import _featColors as featColors
import base64
import io

IPythonConsole.ipython_3d = True
from IPython.display import SVG


def draw_mol(m, feats=None, p=None, confId=-1, hydrogen=True):

    m = Chem.Mol(m)

    if p is None:
        p = py3Dmol.view(width=400, height=400)
    p.removeAllModels()
    if not hydrogen:
        m = Chem.RemoveHs(m)
    IPythonConsole.addMolToView(m, p, confId=confId)

    def colorToHex(rgb):
        """Convert an RGB tuple with values in the range [0, 1] to a hex code."""
        return "0x" + "".join(f"{int(round(x * 255)):02X}" for x in rgb)

    feats = feats or []
    for feat in feats:
        pos = feat.GetPos()
        clr = featColors.get(feat.GetFamily(), (0.5, 0.5, 0.5))
        p.addSphere(
            {
                "center": {"x": pos.x, "y": pos.y, "z": pos.z},
                "radius": 0.5,
                "color": colorToHex(clr),
            }
        )
    p.zoomTo()
    p.show()


def draw_mols(
    ms,
    p=None,
    confId=-1,
    hydrogen=True,
    colors=("cyanCarbon", "redCarbon", "blueCarbon"),
):
    if p is None:
        p = py3Dmol.view(width=400, height=400)
    p.removeAllModels()
    for i, m in enumerate(ms):
        if not hydrogen:
            m = Chem.RemoveHs(m)
        IPythonConsole.addMolToView(m, p, confId=confId)
    for i, m in enumerate(ms):
        p.setStyle(
            {
                "model": i,
            },
            {"stick": {"colorscheme": colors[i % len(colors)]}},
        )
    p.zoomTo()
    p.show()


def draw_grid(
    mols, labels=None, find_mcs=False, align_substructure=True, highlightAtomLists=None
):
    mols = [Chem.MolFromSmiles(mol) if isinstance(mol, str) else mol for mol in mols]

    if not labels:
        labels = [Chem.MolToSmiles(mol) for mol in mols]

    if not find_mcs:
        return Chem.Draw.MolsToGridImage(
            mols, legends=labels, highlightAtomLists=highlightAtomLists
        )

    else:

        res = rdFMCS.FindMCS(mols)
        # mcs_smarts = res.smartsString
        mcs_mol = Chem.MolFromSmarts(res.smartsString)
        smarts = res.smartsString
        smart_mol = Chem.MolFromSmarts(smarts)
        smarts_and_mols = [smart_mol] + mols

        smarts_legend = "Max. substructure match"

        legends = [smarts_legend] + labels

        matches = [""] + [mol.GetSubstructMatch(mcs_mol) for mol in mols]

        subms = [x for x in smarts_and_mols if x.HasSubstructMatch(mcs_mol)]

        AllChem.Compute2DCoords(mcs_mol)

        if align_substructure:
            for m in subms:
                _ = AllChem.GenerateDepictionMatching2DStructure(m, mcs_mol)

        drawing = Chem.Draw.MolsToGridImage(
            smarts_and_mols, highlightAtomLists=matches, legends=legends
        )

        return drawing


def draw_mcs(
    smiles: list[str] or dict[str, str],
    align_substructure: bool = True,
    verbose: bool = False,
    show_mcs=True,
    highlight: bool = True,
    highlight_common: bool = True,
    **kwargs,
):
    """
    Convert a list (or dictionary) of SMILES strings to an RDKit grid image of the maximum common substructure (MCS) match between them

    :returns: RDKit grid image, and (if verbose=True) MCS SMARTS string and molecule, and list of molecules for input SMILES strings
    :rtype: RDKit grid image, and (if verbose=True) string, molecule, and list of molecules
    :param molecules: The SMARTS molecules to be compared and drawn
    :type molecules: List of (SMARTS) strings, or dictionary of (SMARTS) string: (legend) string pairs
    :param align_substructure: Whether to align the MCS substructures when plotting the molecules; default is True
    :type align_substructure: boolean
    :param verbose: Whether to return verbose output (MCS SMARTS string and molecule, and list of molecules for input SMILES strings); default is False so calling this function will present a grid image automatically
    :type verbose: boolean
    """
    mols = [Chem.MolFromSmiles(smile) for smile in smiles]
    res = rdFMCS.FindMCS(mols, **kwargs)
    mcs_smarts = res.smartsString
    mcs_mol = Chem.MolFromSmarts(res.smartsString)
    smarts = res.smartsString
    smart_mol = Chem.MolFromSmarts(smarts)

    if show_mcs:
        smarts_and_mols = [smart_mol] + mols
    else:
        smarts_and_mols = mols

    smarts_legend = "Max. substructure match"

    # If user supplies a dictionary, use the values as legend entries for molecules
    if isinstance(smiles, dict):
        mol_legends = [smiles[molecule] for molecule in smiles]
    else:
        mol_legends = ["" for mol in mols]

    if show_mcs:
        legends = [smarts_legend] + mol_legends
    else:
        legends = mol_legends

    def inverse_indices(mol, indices):
        inverted = []
        for index, _ in enumerate(mol.GetAtoms()):
            if index not in indices:
                inverted.append(index)
        return tuple(inverted)

    if highlight:
        if not highlight_common:
            matches = [
                inverse_indices(mol, mol.GetSubstructMatch(mcs_mol)) for mol in mols
            ]
        else:
            matches = [mol.GetSubstructMatch(mcs_mol) for mol in mols]

        if show_mcs:
            matches = [""] + matches

    subms = [x for x in smarts_and_mols if x.HasSubstructMatch(mcs_mol)]

    AllChem.Compute2DCoords(mcs_mol)

    if align_substructure:
        for m in subms:
            _ = AllChem.GenerateDepictionMatching2DStructure(m, mcs_mol)

    if highlight:
        drawing = Draw.MolsToGridImage(
            smarts_and_mols, highlightAtomLists=matches, legends=legends
        )
    else:
        drawing = Draw.MolsToGridImage(smarts_and_mols, legends=legends)

    if verbose:
        return drawing, mcs_smarts, mcs_mol, mols
    else:
        return drawing


def view_difference(mol1, mol2):
    mcs = rdFMCS.FindMCS([mol1, mol2])
    mcs_mol = Chem.MolFromSmarts(mcs.smartsString)
    match1 = mol1.GetSubstructMatch(mcs_mol)
    target_atm1 = []
    for atom in mol1.GetAtoms():
        if atom.GetIdx() not in match1:
            target_atm1.append(atom.GetIdx())
    match2 = mol2.GetSubstructMatch(mcs_mol)
    target_atm2 = []
    for atom in mol2.GetAtoms():
        if atom.GetIdx() not in match2:
            target_atm2.append(atom.GetIdx())
    return Draw.MolsToGridImage(
        [mol1, mol2], highlightAtomLists=[target_atm1, target_atm2]
    )


def draw_highlighted_mol(
    mol,
    index_color_pairs,
    flat: bool = False,
    indices: bool = False,
    legend=None,
    size=(600, 300),
):
    drawer = rdMolDraw2D.MolDraw2DSVG(*size)

    if flat:
        mol = Chem.Mol(mol)
        conformer_ids = [c.GetId() for c in mol.GetConformers()]
        for c_id in conformer_ids:
            mol.RemoveConformer(c_id)

    if legend is None:
        legend = ""

    if indices:
        for i, atom in enumerate(mol.GetAtoms()):
            atom.SetProp("atomNote", f"{i}")
        # for i, atom in enumerate(mol.GetAtoms()):
        # atom.SetAtomMapNum(atom.GetIdx())

    drawer.DrawMolecule(
        mol,
        highlightAtoms=[x[0] for x in index_color_pairs],
        highlightAtomColors={x[0]: x[1] for x in index_color_pairs},
        legend=legend,
    )
    # drawer.DrawMolecule(mol)#,highlightAtoms=[x[0] for x in index_color_pairs],highlightAtomColors={x[0]:x[1] for x in index_color_pairs})

    drawer.FinishDrawing()
    svg = drawer.GetDrawingText().replace("svg:", "")
    return SVG(svg)


def draw_flat(mol, indices=False, map_nums: bool = False, legend=None, size=(600, 300)):

    # remove conformers
    mol = Chem.Mol(mol)
    conformer_ids = [c.GetId() for c in mol.GetConformers()]
    for c_id in conformer_ids:
        mol.RemoveConformer(c_id)

    # orient to flat
    flat_mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
    AllChem.Compute2DCoords(mol)
    AllChem.Compute2DCoords(flat_mol)
    AllChem.GenerateDepictionMatching2DStructure(mol, flat_mol)

    drawer = rdMolDraw2D.MolDraw2DSVG(*size)

    if indices:
        for i, atom in enumerate(mol.GetAtoms()):
            atom.SetProp("atomNote", f"{i}")
    if map_nums:
        for i, atom in enumerate(mol.GetAtoms()):
            atom.SetAtomMapNum(atom.GetIdx())

    drawer.DrawMolecule(mol)

    drawer.FinishDrawing()
    svg = drawer.GetDrawingText().replace("svg:", "")
    return SVG(svg)


def smiles_to_pngstr(smiles):
    mol = Chem.MolFromSmiles(smiles)
    image = Draw.MolToImage(mol)
    buff = io.BytesIO()
    image.save(buff, format="png")
    return base64.b64encode(buff.getvalue()).decode("utf-8")



from .ndfes import prepare_metafile

# from .ndfes import run_ndfes
# from .ndfes import plot_2d_fes
# from .ndfes import plot_window_pmf
# from .ndfes import generate_pmf_object

from .ndfes import NDFES

from .hijack import FakeAtoms
from .hijack import FakeSurfaceCalculator



def csv_strip(filename, output=None, overwrite=False, comment_chars="#@!"):
    import re
    import mout
    import mcol

    with open(filename, "r") as file:
        lines = file.readlines()
    file.close()

    if output is None and not overwrite:
        mout.errorOut("No output specified", fatal=True)

    if overwrite:
        mout.warningOut("Overwriting " + mcol.file + filename + mcol.warning + "!")
        output = filename

    file = open(output, "w")

    test_line = re.sub(r"\t", " ", lines[0])
    test_line = re.sub(" +", " ", test_line)
    test_line = test_line.strip()

    if test_line != lines[0]:
        for line in lines:

            if line[0] in comment_chars:
                continue

            line = re.sub(r"\t", " ", line)
            line = re.sub(" +", " ", line)
            line = line.strip()
            file.write(line + "\n")
    else:
        mout.warningOut("Skipping already stipped file: " + filename)

    file.close()


def parseDat(
    filename,
    num_columns=2,
    header_rows=1,
    delimiter=" ",
    debug=False,
    pre_strip=False,
    clean_nan=False,
    comment_chars="#@!",
):
    import mout
    import pandas
    import os
    import math

    columns = list(range(num_columns))

    labels = []

    for i in columns:

        if i == 0:
            labels.append("x")
        else:
            labels.append("y" + str(i))

    if pre_strip:
        # csv_strip(filename,overwrite=True)
        csv_strip(
            filename, output="__temp__", overwrite=False, comment_chars=comment_chars
        )

        dataframe = pandas.read_csv(
            "__temp__",
            skiprows=header_rows,
            delimiter=delimiter,
            usecols=columns,
            names=labels,
        )

        os.system("rm __temp__")

    else:

        dataframe = pandas.read_csv(
            filename,
            skiprows=header_rows,
            delimiter=delimiter,
            usecols=columns,
            names=labels,
            comment="#",
        )

    if debug:
        print(dataframe)

    if num_columns > 2:

        if debug:
            mout.headerOut("many")

        big_y = []
        x = dataframe.iloc[:, 0].values

        for i in range(1, num_columns):

            y = list(dataframe.iloc[:, i].values)

            big_y.append(y)

            if debug:
                mout.varOut("big_y len", len(big_y))

        if clean_nan:
            if num_columns > 2:
                new_x = []
                new_ys = []
                for x, ys in zip(x, big_y):

                    # print("DEBUG",x,ys)

                    try:
                        if math.isnan(float(x)):
                            continue
                    except ValueError:
                        continue
                    for y in ys:
                        try:
                            if math.isnan(float(y)):
                                continue
                        except ValueError:
                            continue

                    new_x.append(x)
                    new_ys.append(ys)
                x = new_x
                big_y = new_ys

        return x, big_y

    elif num_columns == 1:

        x = dataframe.iloc[:, 0].values

        if clean_nan:
            new_x = []
            for this_x in list(x):
                if math.isnan(float(this_x)):
                    continue
                new_x.append(this_x)
            return new_x

        return list(x)

    else:

        x = dataframe.iloc[:, 0].values
        y = dataframe.iloc[:, 1].values

        if clean_nan:
            new_x = []
            new_y = []
            for this_x, this_y in zip(list(x), list(y)):
                if math.isnan(float(this_x)):
                    continue
                if math.isnan(float(this_y)):
                    continue
                new_x.append(this_x)
                new_y.append(this_y)
            return new_x, new_y

        return list(x), list(y)



from mwin.curses import CursesApp, Button, Text

import curses

from .group import AtomGroup
from .system import System
from .chain import Chain
from .residue import Residue
from .atom import Atom
from .amino import AminoAcid
from .nucleic import NucleicAcid

""" To Do's

* Simplify chains of single atom/residue types repeated over and over
* Distinguish between hetatm/backbone/sidechain atoms
* Add more capabilities to context windows
* Add cursor selection possibility

"""


def tree(obj):
    """Open a CLI App showing the hierarchical tree of a MolParse System."""

    viewer = TreeViewer(obj)


class TreeViewer(CursesApp):
    """CLI App showing the hierarchical tree of a MolParse System. Use molparse.tree(sys) to start."""

    def __init__(self, obj):

        super(TreeViewer, self).__init__(debug=False)

        if isinstance(obj, str):
            from .io import parse

            obj = parse(obj)

        self._obj = obj

        if isinstance(obj, System):

            obj._expand = True

            if obj.num_chains == 1:
                obj.chains[0]._expand = True

                if obj.chains[0].num_residues == 1:
                    obj.chains[0].residues[0]._expand = True

        if isinstance(obj, AtomGroup):
            obj._expand = True

        else:
            import mout

            mout.errorOut(f"TreeViewer not supported with {self._obj_type}")

        try:

            self.drawtree()
            self.firstdraw()

            while True:

                redraw = self.draw()

                if redraw:
                    self.clear_drawables()
                    self.drawtree()
                    self.drawcore()

        except KeyboardInterrupt:
            pass
        except curses.error as e:
            self.close()
            print()
            import mout

            mout.errorOut("TreeViewer exited due to a curses error.")
            if "prefresh() returned ERR" in str(e):
                mout.error("Maybe you resized the terminal window?")
            mout.error(e)
        except KeyError as e:
            self.close()
            print()
            import mout

            mout.errorOut("TreeViewer exited due to a KeyError (maybe plot3d?).")
            mout.error(e)
        except AttributeError as e:
            self.close()
            print()
            import mout

            mout.error(
                "TreeViewer exited due to an AttributeError (maybe outdated MPyTools?)."
            )
            mout.error(e)
        except Exception as e:
            self.close()
            print()
            import mout

            mout.errorOut("TreeViewer exited due to an error.")
            mout.error(e)

        self.close()

    @property
    def obj(self):
        return self._obj

    def drawtree(self):
        max_index = self.obj.children[-1].index
        if self.type_str(self.obj.children[-1]) not in ["System", "AtomGroup", "Chain"]:
            max_number = self.obj.children[-1].number
        else:
            max_number = None
        self.recursive_tree(self.obj, 0, max_index, max_number, 0)

    def recursive_tree(self, parent, line, max_index, max_number, depth=0):

        line = self.object_line(parent, line, max_index, max_number, depth)

        if parent.children and parent._expand:
            max_index = parent.children[-1].index
            for child in parent.children:
                line = self.recursive_tree(
                    child,
                    line,
                    max_index,
                    max_number,
                    depth + 1,
                )

        return line

    def has_pad_space(self, num_children):
        if self.pad_h - self.max_padline < num_children:
            return False
        return True

    def spawn_context(self, obj, line, col):

        col = self._last_pressed.col

        items, clickables = obj._context_info

        widgets = []

        for i, (key, prop) in enumerate(items.items()):
            label = Text(
                f"{key}: ", line + 1 + i, col, color_pair=self.WHITE_INV | curses.A_BOLD
            )
            data = Text(
                f"{prop} ", line + 1 + i, label.endcol, color_pair=self.WHITE_INV
            )
            widgets.append(label)
            widgets.append(data)

        for j, (key, func) in enumerate(clickables.items()):
            label = Button(
                app=self,
                line=line + 1 + i + j,
                col=col,
                active=False,
                enabler=func,
                target=obj,
                disabler=func,
                color_inactive=self.GREEN_INV,
                color_active=self.GREEN_INV,
                name=key,
            )
            widgets.append(label)

        self.context_menu(line, col, widgets)

    def object_line(self, obj, line, max_index, max_number, depth):

        col = 4 * depth

        obj._tree = self
        obj._gui_line = line

        # add expand/collapse button
        if not isinstance(obj, Atom) and self.has_pad_space(len(obj.children)):

            f1 = lambda x: x.expand()
            f2 = lambda x: x.collapse()

            button = Button(
                app=self,
                line=line,
                col=col,
                active=obj._expand,
                enabler=f1,
                color_active=self.RED_INV,
                disabler=f2,
                target=obj,
                name=f" > ",
                activename=f" v ",
            )
            self.add_button(button)
            col = button.endcol + 1

        # no expand functionality
        else:

            text = Text(f"   ", line, col, bold=False)
            self.add_text(text)
            col = text.endcol + 1

        # alternative sites
        if self.type_str(obj) == "Atom" and obj.alternative_site:
            text = Text(
                f"({obj.alternative_site})",
                line,
                col,
                bold=True,
                color_pair=self.YELLOW,
            )
            self.add_text(text)
            col = text.endcol + 1

        # object type
        text = Text(f"{self.type_str(obj)} ", line, col, bold=True)
        self.add_text(text)

        # object index
        if self.type_str(obj) not in ["System", "AtomGroup"]:
            # index
            width = len(str(max_index))
            text = Text(
                f"{str(obj.index).rjust(width)} ",
                line,
                text.endcol,
                color_pair=self.GREEN,
                bold=True,
            )
            self.add_text(text)

        if self.type_str(obj) not in ["System", "AtomGroup", "Chain"]:
            # number
            width = len(str(max_number))
            text = Text(
                f"{str(obj.number).rjust(width)} ",
                line,
                text.endcol,
                color_pair=self.GREEN,
                bold=True,
            )
            self.add_text(text)

        # object name is also a button:
        if self.type_str(obj) != "Atom":
            name = obj.name

            if len(name) > 23:
                name = f"{name[0:20]}..."

        else:
            name = obj.symbol

        f1 = lambda x: x.spawn_context(line, int(text.endcol) - len(name))
        f2 = lambda x: x.hide_context()

        if not isinstance(obj, Atom):
            # obj._show_context = False

            text = Button(
                app=self,
                line=line,
                col=text.endcol,
                active=obj._show_context,
                enabler=f1,
                color_active=self.CYAN_INV,
                color_inactive=self.CYAN_INV,
                # color_active=self.WHITE_INV|curses.A_BOLD,
                # color_inactive=curses.A_UNDERLINE|self.CYAN|curses.A_BOLD,
                disabler=f2,
                target=obj,
                name=name,
            )
            # activename=f'!{name}!')
            self.add_button(text)

        else:
            text = Text(name, line, text.endcol, color_pair=self.CYAN, bold=True)
            self.add_text(text)

        # residue/chain type
        if self.type_str(obj) not in ["System", "Atom", "AtomGroup"]:
            color_pair = self.color_from_type(obj.type)
            text = Text(f"(", line, text.endcol + 1, bold=True)
            self.add_text(text)
            text = Text(
                f"{obj.type}", line, text.endcol, color_pair=color_pair, bold=True
            )
            self.add_text(text)
            text = Text(f")", line, text.endcol, bold=True)
            self.add_text(text)

        # position
        if self.type_str(obj) in ["Atom"]:
            text = Text(
                f" [{obj.x:7.3f}, {obj.y:7.3f}, {obj.z:7.3f}]", line, text.endcol
            )
            self.add_text(text)

            text = Text(
                f" {obj.name}", line, text.endcol, color_pair=self.MAGENTA, bold=True
            )
            self.add_text(text)

        # extra info
        if self.scr_w >= 80:
            col = self.extra_info(obj, line, text.endcol + 1)
        else:
            col = text.endcol

        # plotting buttons
        if self.type_str(obj) in [
            "System",
            "Chain",
            "Residue",
            "AminoAcid",
            "NucleicAcid",
            "AtomGroup",
        ]:
            col += 1

            f1 = lambda x: x.plot3d()

            button = Button(
                app=self,
                line=line,
                col=col,
                active=False,
                color_inactive=curses.A_UNDERLINE | curses.A_BOLD,
                enabler=f1,
                disabler=f1,
                target=obj,
                name=f"plotly",
            )

            self.add_button(button)
            col = button.endcol + 1

            f1 = lambda x: x.view()

            button = Button(
                app=self,
                line=line,
                col=col,
                active=False,
                color_inactive=curses.A_UNDERLINE | curses.A_BOLD,
                enabler=f1,
                disabler=f1,
                target=obj,
                name=f"ASE",
            )

            self.add_button(button)
            col = button.endcol + 1

        if self.type_str(obj) in ["Residue", "AminoAcid", "NucleicAcid", "Chain"]:
            file_out = f"{self.type_str(obj)}_{obj.name}_{obj.index}.pdb"

            f1 = lambda x: x.write(file_out)

            button = Button(
                app=self,
                line=line,
                col=col,
                active=False,
                color_inactive=curses.A_UNDERLINE | curses.A_BOLD,
                enabler=f1,
                disabler=f1,
                target=obj,
                name=f"PDB",
            )

            self.add_button(button)
            col = button.endcol + 1

        return line + 1

    def extra_info(self, obj, line, col):
        if isinstance(obj, System):
            text = Text(
                f'contains: {self.quantity_str(obj.num_chains, "chain")}, {self.quantity_str(obj.num_residues, "residue")}, {self.quantity_str(obj.num_atoms, "atom")} ',
                line,
                col,
            )
            self.add_text(text)
        elif isinstance(obj, Chain):
            text = Text(
                f'contains: {self.quantity_str(obj.num_residues, "residue")}, {self.quantity_str(obj.num_atoms, "atom")} ',
                line,
                col,
            )
            self.add_text(text)
        elif isinstance(obj, Residue):
            if (
                self.type_str(obj) in ["Residue", "AminoAcid", "NucleicAcid"]
                and obj.contains_alternative_sites
            ):
                try:
                    s = ",".join(sorted(obj.alternative_sites))
                except TypeError:
                    raise TypeError(
                        f"Problem with alternative_sites {obj=} {type(obj)=} {obj.alternative_sites=} "
                    )
                s = f"{s:5}"
                text = Text(f"alt=", line, col, bold=True, color_pair=self.YELLOW)
                self.add_text(text)
                col = text.endcol
                text = Text(f"{s}", line, col, bold=True, color_pair=self.YELLOW)
                self.add_text(text)
                col = text.endcol + 1
                text = Text(f'{self.quantity_str(obj.num_atoms, "atom")} ', line, col)

            else:
                text = Text(
                    f'contains: {self.quantity_str(obj.num_atoms, "atom")} ', line, col
                )

            self.add_text(text)

        else:
            return col
        return text.endcol

    def quantity_str(self, number, name):
        if number > 1:
            return f"{number} {name}s"
        else:
            return f"{number} {name}"

    def type_str(self, obj):
        if isinstance(obj, System):
            return "System"
        if isinstance(obj, Chain):
            return "Chain"
        if isinstance(obj, AminoAcid):
            return "AminoAcid"
        if isinstance(obj, NucleicAcid):
            return "NucleicAcid"
        if isinstance(obj, Residue):
            return "Residue"
        if isinstance(obj, Atom):
            return "Atom"
        if isinstance(obj, AtomGroup):
            return "AtomGroup"

    def color_from_type(self, t):
        match t:
            case "PRO":
                return self.GREEN

            case "DNA":
                return self.RED

            case "SOL":
                return self.WHITE

            case "ION":
                return self.CYAN

            case "LIP":
                return self.YELLOW

            case "LIG":
                return self.MAGENTA

            case "N/A":
                return self.WHITE

            case _:
                return self.WHITE



#!/usr/bin/env python3

import mout
from molparse.io import parse

import mrich as logger


from mout.testing import TestStatus


@mout.debug_log
def check_res_delete_names(system, names):
    test_status = TestStatus()

    system.remove_residues(names=names)

    sys_names = system.res_names
    for name in names:
        if name in sys_names:
            test_status.error(f"Residue(s) with {name=} not deleted")

    return test_status


@mout.debug_log
def check_res_delete_indices(system, indices):
    test_status = TestStatus()

    system.remove_residues(indices=indices)

    sys_indices = system.res_indices
    for index in indices:
        if index in sys_indices:
            test_status.error(f"Residue(s) with {index=} not deleted")

    return test_status


@mout.debug_log
def check_res_delete_numbers(system, numbers):
    test_status = TestStatus()

    system.remove_residues(numbers=numbers)

    sys_numbers = system.res_numbers
    for number in numbers:
        if number in sys_numbers:
            test_status.error(f"Residue(s) with {number=} not deleted")

    return test_status


@mout.debug_log
def check_atom_delete_names(system, names):
    test_status = TestStatus()

    system.remove_atoms(names=names)

    sys_names = system.atom_names
    for name in names:
        if name in sys_names:
            test_status.error(f"Residue(s) with {name=} not deleted")

    return test_status


@mout.debug_log
def check_atom_delete_indices(system, indices):
    test_status = TestStatus()

    system.remove_atoms(indices=indices)

    sys_indices = system.atom_indices
    for index in indices:
        if index in sys_indices:
            test_status.error(f"Residue(s) with {index=} not deleted")

    return test_status


@mout.debug_log
def check_atom_delete_numbers(system, numbers):
    test_status = TestStatus()

    system.remove_atoms(numbers=numbers)

    sys_numbers = system.atom_numbers
    for number in numbers:
        if number in sys_numbers:
            test_status.error(f"Residue(s) with {number=} not deleted")

    return test_status


def main():

    # get a test system
    import urllib.request

    entry = "3PJR"

    logger.print(f"Retrieving {entry} from protein data bank...")

    urllib.request.urlretrieve(
        f"http://files.rcsb.org/download/{entry}.pdb", f"{entry}.pdb"
    )

    sys = parse(f"{entry}.pdb")

    sys.summary()

    check_res_delete_names(sys, ["DT3", "DG", "TYR"])
    check_res_delete_indices(sys, [1, 3, 50, 300, 1000, 2000])
    check_res_delete_numbers(sys, [1, 3, 50, 300, 1000, 2000])

    check_atom_delete_names(sys, ["H41", "N9", "CG3"])
    check_atom_delete_indices(sys, [1, 3, 50, 300, 1000, 2000, 8000, 20000])
    check_atom_delete_numbers(sys, [1, 3, 50, 300, 1000, 2000, 8000, 20000])


if __name__ == "__main__":
    main()
